#
# This file is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it 
#   under the terms of the GNU Lesser General Public License as published 
#   by the Free Software Foundation, either version 3 of the License, or 
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but 
#   WITHOUT ANY WARRANTY; without even the implied warranty of 
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU 
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public 
#   License along with Dragonfly.  If not, see 
#   <http://www.gnu.org/licenses/>.
#

"""
Engine class for CMU Pocket Sphinx
"""

import contextlib
import logging
import os
import time
import wave

from threading import Timer, RLock, current_thread

from dragonfly import Grammar, Window
from dragonfly.engines.backend_sphinx import is_engine_available
from dragonfly.engines.backend_sphinx.training import TrainingDataWriter

from .dictation import SphinxDictationContainer
from .recobs import SphinxRecObsManager
from ..base import EngineBase, EngineError, MimicFailure

try:
    from sphinxwrapper import *
    from jsgf import GrammarError, RootGrammar, PublicRule, Literal
    from pyaudio import PyAudio

    from .dragonfly2jsgf import Translator
    from .compiler import SphinxJSGFCompiler
    from .grammar_wrapper import GrammarWrapper, ProcessingState
except ImportError:
    # Import a few things here optionally for readability (the engine won't start
    # without them) and so that autodoc can import this module without them.
    pass


class UnknownWordError(Exception):
    pass


class SphinxEngine(EngineBase):
    """Speech recognition engine back-end for CMU Pocket Sphinx."""

    _name = "sphinx"
    DictationContainer = SphinxDictationContainer

    # A hypothesis from Pocket Sphinx can be None, so this object used to indicate
    # when a dictation hypothesis has not been calculated for the current utterance.
    _DICTATION_HYP_UNSET = object()

    def __init__(self):
        EngineBase.__init__(self)

        # Set up the engine logger
        logging.basicConfig()

        try:
            import sphinxwrapper
            import jsgf
            import pyaudio
        except ImportError:
            self._log.error("%s: Failed to import jsgf, pyaudio and/or "
                            "sphinxwrapper. Are they installed?" % self)
            raise EngineError("Failed to import Pocket Sphinx engine dependencies.")

        # Import and set the default configuration module. This can be changed later
        # using the config property.
        from . import config
        self._config = None
        self.config = config

        # Set other variables
        self._decoder = None
        self._audio_buffers = []
        self.compiler = SphinxJSGFCompiler()
        self._recognition_observer_manager = SphinxRecObsManager(self)
        self._in_progress_states = set()
        self._current_grammar_wrapper = None
        self._keyphrase_thresholds = {}
        self._keyphrase_functions = {}

        # Set up keyphrase search names and valid search names for grammars.
        self._keyphrase_search_names = {"_key_phrases", "_wake_phrase"}
        self._valid_searches = set()

        # Members used in recognition timeouts
        self._last_recognition_time = None
        self._timeout_timer = None
        self._thread_lock = RLock()

        # Recognising loop control variables
        self._recognising = False
        self._cancel_recognition_next_time = False
        self._last_recognition_time = None

        # Variable used in caching dictation hypotheses for speech utterances.
        self._current_dictation_hyp = self._DICTATION_HYP_UNSET

        # Training data writer, initialised on connect().
        self._training_data_writer = None

        # Variable used in training sessions.
        self._training_session_active = False

    @property
    def config(self):
        """
        Python module/object containing engine configuration.

        Setting this property will raise an :class:`EngineError` if the given
        configuration object doesn't define each configuration option.

        You will need to restart the engine with :meth:`disconnect` and
        :meth:`connect` if you have made configuration changes with this property
        after :meth:`connect` has been called.

        :raises: EngineError
        :returns: config module/object
        """
        return self._config

    @config.setter
    def config(self, value):
        # Validate configuration module
        self.validate_config(value)
        self._config = value

    @staticmethod
    def validate_config(engine_config):
        attributes = [
            "DECODER_CONFIG", "PYAUDIO_STREAM_KEYWORD_ARGS", "LANGUAGE",
            "NEXT_PART_TIMEOUT", "START_ASLEEP", "WAKE_PHRASE",
            "WAKE_PHRASE_THRESHOLD", "SLEEP_PHRASE", "SLEEP_PHRASE_THRESHOLD",
            "TRAINING_DATA_DIR", "START_TRAINING_PHRASE",
            "START_TRAINING_THRESHOLD", "END_TRAINING_PHRASE",
            "END_TRAINING_THRESHOLD"
        ]
        not_preset = []
        for attr in attributes:
            if not hasattr(engine_config, attr):
                not_preset.append(attr)

        if not_preset:
            # Raise an error with the attributes that weren't set.
            not_preset.sort()
            raise EngineError("invalid engine configuration. The following "
                              "attributes were not present: %s"
                              % ", ".join(not_preset))

    @property
    def observer_manager(self):
        return self._recognition_observer_manager

    @property
    def log(self):
        return self._log

    def connect(self):
        """
        Set up the CMU Pocket Sphinx decoder.

        This method does nothing if the engine is already connected.
        """
        if self._decoder:
            return

        # Check that no search argument other than -lm is specified, as the
        # engine will be handling that side of things.
        decoder_config = self.config.DECODER_CONFIG
        search_args = search_arguments_set(decoder_config)

        # Note: len(search_args) > 1 is handled by sphinxwrapper internally by
        # raising an error, so we don't need handle that here.
        if len(search_args) == 1 and search_args[0] != "-lm":
            raise EngineError("invalid PS configuration: please do not specify "
                              "'%s' in the Config object" % search_args[0])

        # Initialise a new decoder with the given configuration
        self._decoder = PocketSphinx(decoder_config)
        self._valid_searches.add(self.dictation_search_name)

        # Set up callback function wrappers
        def hypothesis(hyp):
            # Set speech to the hypothesis string or None if there isn't one
            speech = hyp.hypstr if hyp else None
            return self._hypothesis_callback(speech, False)

        def speech_start():
            return self._speech_start_callback(False)

        self._decoder.hypothesis_callback = hypothesis
        self._decoder.speech_start_callback = speech_start

        try:
            # Check that all words in configuration keyphrases are in the
            # pronunciation dictionary. An UnknownWordError will be raised if one
            # isn't.
            words = []
            words.extend(self.config.WAKE_PHRASE.split())
            words.extend(self.config.SLEEP_PHRASE.split())
            words.extend(self.config.START_TRAINING_PHRASE.split())
            words.extend(self.config.END_TRAINING_PHRASE.split())
            self._validate_words(words, "keyphrase")

            # Set up wake phrase search using the engine configuration.
            self._decoder.set_kws_list("_wake_phrase", {
                self.config.WAKE_PHRASE: self.config.WAKE_PHRASE_THRESHOLD
            })

            # Add the sleep keyphrase + threshold to call pause_recognition when
            # heard.
            self.set_keyphrase(
                self.config.SLEEP_PHRASE, self.config.SLEEP_PHRASE_THRESHOLD,
                self.pause_recognition
            )

            # Add the start/end training session phrases, thresholds and methods
            # to call when heard.
            self.set_keyphrase(
                self.config.START_TRAINING_PHRASE,
                self.config.START_TRAINING_THRESHOLD,
                self.start_training_session
            )
            self.set_keyphrase(
                self.config.END_TRAINING_PHRASE,
                self.config.END_TRAINING_THRESHOLD,
                self.end_training_session
            )
        except UnknownWordError as e:
            # Log the error about unknown words.
            self._log.error(e)

        # Initialise a TrainingDataWriter instance if specified.
        if self.config.TRAINING_DATA_DIR:
            # Get required audio configuration from the engine config.
            channels, sample_width, rate = (
                self.config.PYAUDIO_STREAM_KEYWORD_ARGS["channels"],
                PyAudio().get_sample_size(
                    self.config.PYAUDIO_STREAM_KEYWORD_ARGS["format"]
                ),
                self.config.PYAUDIO_STREAM_KEYWORD_ARGS["rate"],
            )
            self._training_data_writer = TrainingDataWriter(
                self.config.TRAINING_DATA_DIR, "training",
                channels, sample_width, rate
            )
        else:
            self._training_data_writer = None

    def _free_engine_resources(self):
        """
        Internal method for freeing the resources used by the engine.
        """
        # Cancel and free the timeout timer if it is active.
        self._cancel_and_free_timeout_timer()

        # Free the decoder and clear the audio buffer list
        self._decoder = None
        self._audio_buffers = []

        # Reset other variables
        self._cancel_recognition_next_time = False
        self._current_grammar_wrapper = None
        self._current_dictation_hyp = self._DICTATION_HYP_UNSET
        self._training_session_active = False

        # Close the training data files and discard the writer if necessary.
        if self._training_data_writer:
            self._training_data_writer.close_files()
            self._training_data_writer = None

        # Clear dictionaries and sets
        self._grammar_wrappers.clear()
        self._in_progress_states.clear()
        self._valid_searches.clear()
        self._keyphrase_thresholds.clear()
        self._keyphrase_functions.clear()

    def disconnect(self):
        """
        Deallocate the CMU Sphinx decoder and any other resources used by it.

        This method effectively unloads all loaded grammars and key phrases.
        """
        # Free resources if the decoder isn't currently being used to recognise,
        # otherwise stop the recognising loop, which will free the resources safely.
        if not self.recognising:
            self._free_engine_resources()
        else:
            self._recognising = False

    # -----------------------------------------------------------------------
    # Methods for working with grammars.

    def check_valid_word(self, word):
        """
        Check if a word is in the current Sphinx pronunciation dictionary.

        This will always return False if :meth:`connect` hasn't been called.

        :rtype: bool
        """
        if self._decoder:
            return bool(self._decoder.lookup_word(word))

        return False

    def _validate_words(self, words, search_type):
        unknown_words = []

        # Use 'set' to de-duplicate the 'words' list.
        for word in set(words):
            if not self.check_valid_word(word):
                unknown_words.append(word)

        if unknown_words:
            # Sort the word list before using it.
            unknown_words.sort()
            raise UnknownWordError("%s used words not found in the pronunciation "
                                   "dictionary: %s"
                                   % (search_type, ", ".join(unknown_words)))

    def _build_grammar_wrapper(self, grammar):
        return GrammarWrapper(grammar, self)

    def set_grammar(self, wrapper):
        # Set and/or switch to an appropriate Pocket Sphinx search for a given
        # GrammarWrapper.
        # This method is thread safe.
        with self._thread_lock:
            self._set_grammar(wrapper)

    def _set_grammar(self, wrapper):
        self._current_grammar_wrapper = wrapper

        # Return if the engine isn't available or if wrapper is None.
        if not (is_engine_available() and wrapper):
            return

        # Connect to the engine if it isn't connected already.
        self.connect()

        # Don't allow grammar wrappers to use key phrase search names.
        if wrapper.search_name in self._keyphrase_search_names:
            raise EngineError("grammar cannot use '%s' as a search name"
                              % wrapper.search_name)

        # Check if the wrapper's search_name is valid
        if wrapper.search_name in self._valid_searches:
            # If the wrapper's search_name is not the active search name, then swap
            # to it.
            if wrapper.search_name != self._decoder.active_search:
                self._decoder.end_utterance()
                self._decoder.active_search = wrapper.search_name

            # wrapper.search_name is active and is a valid search, so return.
            return

        # Compile and set the jsgf search. If it compiles to nothing or if there is
        # a GrammarError during compilation, fallback to the dictation search
        # instead.
        try:
            compiled = wrapper.dictation_grammar.compile_as_root_grammar()
            if "<root>" not in compiled:
                compiled = ""
        except GrammarError:
            compiled = ""

        if compiled:
            try:
                # Check that each word in the grammar is in the pronunciation
                # dictionary. This will raise an UnknownWordError if one or more
                # aren't.
                self._validate_words(wrapper.grammar_words,
                                     "grammar '%s'" % wrapper.grammar.name)

                # Set the JSGF search
                self._decoder.end_utterance()
                self._decoder.set_jsgf_string(wrapper.search_name, compiled)
                self._decoder.active_search = wrapper.search_name
            except RuntimeError:
                self._log.error("error setting PS JSGF search: %s" % compiled)

                # Return early to avoid adding this search as valid.
                return
        else:
            self._set_dictation_search()

        # Grammar search has been loaded, add the search name to the set.
        self._valid_searches.add(self._decoder.active_search)

    def unset_search(self, name):
        # Unset a Pocket Sphinx search with the given name.
        # Note that this method will not unset the dictation or keyphrase searches.
        # This method is thread safe.
        with self._thread_lock:
            if (name == self.dictation_search_name or
                    name in self._keyphrase_search_names):
                return

            if name in self._valid_searches:
                # TODO Unset the Pocket Sphinx search.
                # Unfortunately, the C function for doing this (ps_unset_search) is
                # not exposed... >_<
                # With the current Python API, this could be done either by freeing
                # the decoder used or by overriding the Pocket Sphinx search with
                # something else.

                # Remove the search from the valid searches set.
                self._valid_searches.remove(name)

    # TODO Add optional context parameter
    def set_keyphrase(self, keyphrase, threshold, func):
        """
        Add a keyphrase to listen for.

        Key phrases take precedence over grammars as they are processed first.
        They cannot be set for specific contexts (yet).

        :param keyphrase: keyphrase to add.
        :param threshold: keyphrase threshold value to use.
        :param func: function or method to call when the keyphrase is heard.
        :type keyphrase: str
        :type threshold: float
        :type func: callable
        """
        # Check that all words in the keyphrase are in the pronunciation dictionary.
        try:
            self._validate_words(keyphrase.split(), "keyphrase")
        except UnknownWordError as e:
            # Log an error and return early.
            self._log.error(e)
            return

        # Add parameters to the relevant dictionaries.
        self._keyphrase_thresholds[keyphrase] = threshold
        self._keyphrase_functions[keyphrase] = func

        # Set the keyphrase search (again)
        self._decoder.end_utterance()
        self._decoder.set_kws_list("_key_phrases", self._keyphrase_thresholds)

    def unset_keyphrase(self, keyphrase):
        """
        Remove a set keyphrase so that the engine no longer listens for it.

        :param keyphrase: keyphrase to remove.
        :type keyphrase: str
        """
        # Remove parameters from the relevant dictionaries. Don't raise an error
        # if there is no such keyphrase.
        self._keyphrase_thresholds.pop(keyphrase, None)
        self._keyphrase_functions.pop(keyphrase, None)

        # Set the keyphrase search (again)
        self._decoder.end_utterance()
        self._decoder.set_kws_list("_key_phrases", self._keyphrase_thresholds)

    def _set_dictation_search(self):
        # Change the active search to the one used for recognising dictation.
        if not self.recognising_dictation:
            self._decoder.end_utterance()  # ensure we're not processing
            self._decoder.active_search = "_default"

    def _load_grammar(self, grammar):
        """ Load the given *grammar* and return a wrapper. """
        self._log.debug("Engine %s: loading grammar %s."
                        % (self, grammar.name))

        grammar.engine = self
        # Dependency checking.
        memo = []
        for r in grammar.rules:
            for d in r.dependencies(memo):
                grammar.add_dependency(d)

        wrapper = self._build_grammar_wrapper(grammar)
        try:
            self.set_grammar(wrapper)
        except UnknownWordError as e:
            # Unknown words should be logged as plain error messages, not
            # exception stack traces.
            self._log.error(e)
        except Exception as e:
            self._log.exception("Failed to load grammar %s: %s."
                                % (grammar, e))
        return wrapper

    def _unload_grammar(self, grammar, wrapper):
        try:
            # Unload the current and default searches for the grammar.
            # It doesn't matter if the names are the same.
            self.unset_search(wrapper.search_name)
            self.unset_search(wrapper.default_search_name)
        except Exception as e:
            self._log.exception("Failed to unload grammar %s: %s."
                                % (grammar, e))

    def set_exclusiveness(self, grammar, exclusive):
        try:
            # Not sure what this is, just going to silently ignore it for now
            self._log.debug("set_exclusiveness called for grammar %s. Not doing "
                            "anything yet.")
        except Exception as e:
            self._log.exception("Engine %s: failed set exclusiveness: %s."
                                % (self, e))

    def activate_grammar(self, grammar):
        self._log.debug("Activating grammar %s." % grammar.name)
        grammar.enable()

    def deactivate_grammar(self, grammar):
        self._log.debug("Deactivating grammar %s." % grammar.name)
        grammar.disable()

    def activate_rule(self, rule, grammar):
        self._log.debug("Activating rule %s in grammar %s." % (rule.name, grammar.name))
        wrapper = self._get_grammar_wrapper(grammar)
        if not wrapper:
            return
        try:
            wrapper.dictation_grammar.enable_rule(rule.name)
            self.unset_search(wrapper.search_name)
            self.set_grammar(wrapper)
        except UnknownWordError as e:
            self._log.error(e)
        except Exception as e:
            self._log.exception("Failed to activate grammar %s: %s."
                                % (grammar, e))

    def deactivate_rule(self, rule, grammar):
        self._log.debug("Deactivating rule %s in grammar %s." % (rule.name, grammar.name))
        wrapper = self._get_grammar_wrapper(grammar)
        if not wrapper:
            return
        try:
            wrapper.dictation_grammar.disable_rule(rule.name)
            self.unset_search(wrapper.search_name)
            self.set_grammar(wrapper)
        except UnknownWordError as e:
            self._log.error(e)
        except Exception as e:
            self._log.exception("Failed to activate grammar %s: %s."
                                % (grammar, e))

    def update_list(self, lst, grammar):
        wrapper = self._get_grammar_wrapper(grammar)
        name = lst.name
        if not wrapper:
            return

        assert isinstance(wrapper, GrammarWrapper)

        # Unfortunately there is no way to update lists for Pocket Sphinx without
        # reloading the grammar, so we'll update the list's JSGF rule and reload.

        # Rebuild the list's JSGF rule using dragonfly2jsgf's Translator class
        new_rule = Translator.translate_list(lst)

        # Find the old list rule in the grammar and modify its expansion
        for r in wrapper.dictation_grammar.rules:
            if r.name == name:
                r.expansion.children = new_rule.expansion.children
                break

        # Reload the grammar's Pocket Sphinx JSGF searches.
        self.unset_search(wrapper.search_name)
        self.unset_search(wrapper.default_search_name)
        self.set_grammar(wrapper)

    # -----------------------------------------------------------------------
    # Miscellaneous methods.

    @property
    def recognising(self):
        """
        Whether the engine is recognising speech in a loop.

        To stop recognition, use :meth:`disconnect`.

        :rtype: bool
        """
        return self._recognising

    @property
    def dictation_search_name(self):
        # The name of the Pocket Sphinx search used for processing speech as
        # dictation.
        return "_default"

    @property
    def recognising_dictation(self):
        # Return whether the engine is currently processing speech as dictation.
        return self._decoder.active_search == self.dictation_search_name

    def get_dictation_hypothesis(self):
        # Get a dictation hypothesis by reprocessing the audio buffers.

        # This method will calculate the dictation hypothesis once and use the same
        # value for successive calls in the processing of a speech utterance.
        if self._current_dictation_hyp is not self._DICTATION_HYP_UNSET:
            return self._current_dictation_hyp

        # Save the current Pocket Sphinx search
        current_search = self._decoder.active_search

        # Switch to the dictation search and reprocess the audio buffers.
        self._set_dictation_search()
        dict_hypothesis = self._decoder.batch_process(self._audio_buffers,
                                                      use_callbacks=False)
        if dict_hypothesis:
            dict_hypothesis = dict_hypothesis.hypstr

        # Switch back to the last search
        if current_search != self._decoder.active_search:
            self._decoder.active_search = current_search

        # Set _current_dictation_hyp
        self._current_dictation_hyp = dict_hypothesis
        return dict_hypothesis

    def _clear_in_progress_states_and_reset(self):
        # While holding the thread lock, clear in-progress states and reset
        # grammar wrappers. This is required as the in-progress state set is
        # a shared resource. This should make the method thread safe.
        with self._thread_lock:
            for wrapper in self._grammar_wrappers.values():
                wrapper.reset_all_sequence_rules()
            self._in_progress_states.clear()

    def _cancel_and_free_timeout_timer(self):
        """
        Cancel and free the timer used for recognition timeouts.
        """
        if not self._timeout_timer:
            return

        # This cannot be done from the Timer thread.
        if current_thread() is self._timeout_timer:
            return

        # Cancel the timer.
        self._timeout_timer.cancel()

        # If the timer is alive, call join on it, timing out after 100ms to prevent
        # deadlocks.
        if self._timeout_timer.isAlive():
            self._timeout_timer.join(0.1)

        # Free it - Timer objects cannot be reused.
        self._timeout_timer = None

    def _on_next_part_timeout(self):
        """
        Method called when there is a next rule part timeout.
        """
        # Do nothing if the timeout value is 0.
        if not self.config.NEXT_PART_TIMEOUT:
            return

        # Execute the following while holding _thread_lock to prevent race
        # conditions, such as the _in_progress_states set changing during
        # iteration and raising an error.
        with self._thread_lock:
            current_time = time.time()
            self._log.info("Recognition time-out after %f seconds"
                           % (current_time - self._last_recognition_time))

            # Remove any states with out of context grammars.
            fg_window = Window.get_foreground()
            for state in tuple(self._in_progress_states):
                # Do context checking for each state.
                state.wrapper.process_begin(fg_window)

                # Remove the state if the grammar is not active.
                if not state.wrapper.grammar_active:
                    self._in_progress_states.remove(state)

            # Calculate the best in-progress state with complete rules if there is
            # one.
            best_state = self._get_best_processing_state(
                [s for s in self._in_progress_states if s.complete_rules]
            )

            if best_state and best_state.wrapper.grammar_active:
                best_state.process(
                    timed_out=True,
                    notify_only=self._training_session_active
                )

            self._clear_in_progress_states_and_reset()

    def _handle_recognition_timeout(self):
        """
        Internal method for handling recognition timeouts. If a recognition timeout
        has not occurred yet, this method will cancel any scheduled timeout.

        The timeout period is specific to rules containing Dictation elements which
        must be recognised in sequence, although if there are matching normal rules,
        those will be processed on timeout.
        """
        next_part_timeout = self.config.NEXT_PART_TIMEOUT
        current_time = time.time()
        if not self._last_recognition_time:
            timed_out = False
        else:
            timed_out = next_part_timeout and (
                current_time >= self._last_recognition_time + next_part_timeout
            )

        if not timed_out:
            # If there is a timer, cancel and free it.
            # It is important that this is not called while holding the lock as
            # a deadlock can occur.
            if self._timeout_timer:
                self._cancel_and_free_timeout_timer()
        else:
            self._clear_in_progress_states_and_reset()

    def _get_best_processing_state(self, states):
        """
        Internal method to get the best ProcessingState object from a list of
        states.

        :type states: list
        :return: ProcessingState
        """
        # Filter out any states that are not processable.
        states = [s for s in states if s.is_processable]
        if not states:
            return None  # no best processing state

        final_states = None

        # Filter the states in 3 ways: states with matching normal rules, states
        # with matching and complete sequence rules, and states with in-progress
        # rules.
        complete_normal = [s for s in states if s.complete_normal_rules]
        complete_seq = [s for s in states if s.complete_sequence_rules]
        in_progress = [s for s in states if s.in_progress_rules]

        # Handle states with complete sequence rules first.
        if complete_seq and self._in_progress_states:
            # Set the final_states list
            final_states = complete_seq

        elif complete_seq and not self._in_progress_states:
            # Set the final_states list
            final_states = complete_seq + complete_normal

        # Then states with matching in-progress rules.
        elif in_progress and not self._in_progress_states:
            # Add states with in-progress and complete normal rules to the engine's
            # set. Complete normal rules should be added in case a next rule part
            # timeout occurs, where they will be processed, if the grammar is still
            # in-context and loaded.
            for state in in_progress:
                self._in_progress_states.add(state)
            for state in complete_normal:
                self._in_progress_states.add(state)

            # Set the final_states list
            final_states = in_progress + complete_normal

        elif in_progress:
            # Replace the old in-progress states with the new ones.
            self._in_progress_states = set(in_progress)

            # Set the final_states list
            final_states = in_progress

        # Then states with only normal matching rules.
        elif complete_normal:
            # Set the final_states list
            final_states = complete_normal

        if final_states:
            # Do any post collection tasks for all states.
            for state in states:
                state.do_post_collection_tasks()

            # Sort the list by the best state to use and return the first state.
            final_states.sort()
            return final_states[0]
        else:
            # Otherwise return None (no best state).
            return None

    def _get_best_hypothesis(self, hypothesises):
        """
        Take a list of speech hypothesises and return the most likely one.

        :type hypothesises: iterable
        :return: str | None
        """
        # Get all distinct, non-null hypothesises using a set and a filter.
        distinct = tuple([h for h in set(hypothesises) if bool(h)])
        if not distinct:
            return None
        elif len(distinct) == 1:
            return distinct[0]  # only one choice

        # Decide between non-null hypothesises using a Pocket Sphinx search with
        # each hypothesis as a grammar rule.
        grammar = RootGrammar()
        grammar.language_name = self.language
        for i, hypothesis in enumerate(distinct):
            grammar.add_rule(PublicRule("rule%d" % i, Literal(hypothesis)))

        compiled = grammar.compile_as_root_grammar()
        name = "_temp"

        # Store the current search name.
        original = self._decoder.active_search

        # Note that there is no need to validate words in this case because each
        # literal in the _temp grammar came from a Pocket Sphinx hypothesis.
        self._decoder.end_utterance()
        self._decoder.set_jsgf_string(name, compiled)
        self._decoder.active_search = name

        # Do the processing.
        result = self._decoder.batch_process(
            self._audio_buffers,
            use_callbacks=False
        )

        # Switch back to the previous search.
        self._decoder.end_utterance()  # just in case
        self._decoder.active_search = original

        if result:
            result = result.hypstr
        return result

    def _set_current_grammar_search(self):
        # Set _current_grammar_wrapper and the active search to the search of an
        # active grammar. If there is no active grammar, set current wrapper to None
        # and active search to the dictation search.
        active_wrappers = list(filter(
            lambda w: w.grammar_active,
            self._grammar_wrappers.values()
        ))
        if active_wrappers:
            # Switch to the first active grammar wrapper's search.
            self._current_grammar_wrapper = active_wrappers[0]
            self.set_grammar(self._current_grammar_wrapper)
        else:
            # No wrapper is usable. Set the dictation search.
            self._current_grammar_wrapper = None
            self._set_dictation_search()

    def _speech_start_callback(self, mimicking):
        # Get context info. Dragonfly has a handy static method for this:
        fg_window = Window.get_foreground()

        # Call process_begin for all grammars so that any out of context grammar
        # will not be used.
        for wrapper in self._grammar_wrappers.values():
            wrapper.process_begin(fg_window)

        # If there are any in-progress states whose grammars are out of context,
        # remove them from the set. Do this while holding the lock to prevent race
        # conditions; the timeout timer could be active at this point.
        with self._thread_lock:
            for state in tuple(self._in_progress_states):
                assert isinstance(state, ProcessingState)
                if not state.wrapper.grammar_active:
                    self._in_progress_states.remove(state)
                    state.wrapper.reset_all_sequence_rules()

        # Handle recognition timeout where speech started too late
        self._handle_recognition_timeout()

        # Set a new current grammar wrapper if necessary.
        if (not self._current_grammar_wrapper or
                not self._current_grammar_wrapper.grammar_active):
            self._set_current_grammar_search()

        # Do a few things with audio buffers if not mimicking.
        if not mimicking:
            # Reprocess current audio buffers if necessary; <decoder>.end_utterance
            # can be called by the above code or by the timer thread.
            if self._decoder.utt_ended and not mimicking:
                self._decoder.batch_process(self._audio_buffers, use_callbacks=False)

            # Trim excess audio buffers from the start of the list. Keep a maximum 3
            # seconds of silence before speech start was detected. This should help
            # increase the performance of batch reprocessing later.
            chunk = self.config.PYAUDIO_STREAM_KEYWORD_ARGS["frames_per_buffer"]
            rate = self.config.PYAUDIO_STREAM_KEYWORD_ARGS["rate"]
            seconds = 3
            n_buffers = int(rate / chunk * seconds)
            self._audio_buffers = self._audio_buffers[-1 * n_buffers:]

            # Add trimmed pre-speech audio buffers to training .wav file.
            if self._training_data_writer:
                for buf in self._audio_buffers:
                    self._training_data_writer.write_to_wave_file(buf)

        # Notify observers
        self.observer_manager.notify_begin()

    def _hypothesis_callback(self, speech, mimicking):
        """
        Internal Pocket Sphinx hypothesis callback method. Calls _process_hypothesis
        and does post-processing afterwards.
        :param speech: hypothesis string | None
        :param mimicking:  whether to treat speech as mimicked speech.
        :rtype: bool
        """
        # Process speech. We should get back a boolean for whether processing
        # occurred as well as the final speech hypothesis.
        processing_occurred, speech = self._process_hypothesises(
            speech, mimicking
        )

        # Do some things that should be done if no processing occurred.
        if not processing_occurred:
            self.observer_manager.notify_failure()
            self._clear_in_progress_states_and_reset()
            self._cancel_and_free_timeout_timer()
            self.set_grammar(self._current_grammar_wrapper)

        # Finalise the current training data and open a new wav file. If speech is
        # None, the current .wav file will be discarded.
        if self._training_data_writer and not mimicking:
            self._training_data_writer.finalise(speech)
            self._training_data_writer.open_next_wav_file()

        # Clear the internal audio buffer list because it is no longer needed.
        self._audio_buffers = []

        # Unset the dictation hypothesis because it is now invalid.
        self._current_dictation_hyp = self._DICTATION_HYP_UNSET

        # Return whether processing occurred in case this method was called by
        # mimic.
        return processing_occurred

    def _process_key_phrases(self, speech, mimicking):
        """
        Processing key phrase searches and return the matched keyphrase (if any).

        :type speech: str
        :param mimicking: whether to treat speech as mimicked speech.
        :rtype: str
        """
        key_phrases_search = "_key_phrases"
        if (self._decoder.active_search == key_phrases_search
                and not speech or not speech and mimicking):
            return None  # no matches

        if self._decoder.active_search != key_phrases_search and not mimicking:
            # End the utterance if it isn't already ended.
            self._decoder.end_utterance()

            # Reprocess using the key phrases search
            last = self._decoder.active_search
            self._decoder.active_search = key_phrases_search
            hyp = self._decoder.batch_process(self._audio_buffers,
                                              use_callbacks=False)

            # Get the hypothesis string.
            speech = hyp.hypstr if hyp else None

            # Restore last search.
            self._decoder.end_utterance()
            self._decoder.active_search = last

        if not speech:
            return None

        # Handle multiple matching key phrases. This appears to be a quirk of how
        # Pocket Sphinx 'kws' searches work. Get the best match instead if this is
        # the case.
        recognised_phrases = speech.split("  ")
        if len(recognised_phrases) > 1:
            # Remove trailing space from the last phrase.
            recognised_phrases[len(recognised_phrases) - 1].rstrip()
            speech = self._get_best_hypothesis(recognised_phrases)
        else:
            speech = speech.rstrip()  # remove trailing whitespace.

        # Call the registered function if there was a match and the function
        # is callable.
        func = self._keyphrase_functions.get(speech, None)
        if callable(func):
            func()

        # Return speech if it matched a keyphrase.
        result = speech if speech in self._keyphrase_functions else None
        return result

    def _process_hypothesises(self, speech, mimicking):
        """
        Internal method to process speech hypothesises. This should only be called
        from 'SphinxEngine._hypothesis_callback' because that method does important
        post processing.

        :type speech: str
        :param mimicking: whether to treat speech as mimicked speech.
        :rtype: tuple
        """
        # Check key phrases search first.
        keyphrase = self._process_key_phrases(speech, mimicking)
        if keyphrase:
            # Keyphrase search matched. Notify observers and return True.
            words_list = [(word, 0) for word in keyphrase.split()]
            self.observer_manager.notify_recognition(words_list)
            return True, keyphrase

        # Otherwise do grammar processing.
        processing_occurred = False
        hypothesises = {}

        # If the engine is processing in-progress rules, only use active in-progress
        # wrappers.
        in_progress_states = self._in_progress_states
        if in_progress_states:
            wrappers = [s.wrapper for s in in_progress_states
                        if s.wrapper.grammar_active]

        # Otherwise collect each active grammar's GrammarWrapper.
        else:
            wrappers = [w for w in self._grammar_wrappers.values()
                        if w.grammar_active]

        self._last_recognition_time = time.time()

        # No grammar has been loaded.
        if not self._current_grammar_wrapper or not wrappers:
            # TODO What should we do here? Output formatted Dictation like DNS?
            return processing_occurred, speech

        if (self._current_grammar_wrapper.search_name !=
                self._decoder.active_search):
            # The grammar's search isn't set, so reprocess speech.
            self.set_grammar(self._current_grammar_wrapper)

            if not mimicking:
                hyp = self._decoder.batch_process(
                    self._audio_buffers,
                    use_callbacks=False
                )
                if hyp and not isinstance(hyp, str):
                    speech = hyp.hypstr
                else:
                    speech = hyp

        if mimicking == "as words":
            # Set dictation hypothesis to None if mimicking with engine.mimic
            self._current_dictation_hyp = None
        elif mimicking:
            # Or if using mimic_phrases, use speech.
            self._current_dictation_hyp = speech

        # Use speech as the hypothesis for the current grammar wrapper.
        hypothesises[self._current_grammar_wrapper.search_name] = speech

        # Batch process audio buffers for each active grammar. Store each
        # hypothesis.
        for wrapper in wrappers:
            if mimicking:
                # Just use speech for every hypothesis if mimicking.
                hypothesises[wrapper.search_name] = speech
            elif wrapper.search_name not in hypothesises.keys():
                # Switch to the search for this grammar if necessary
                if self._decoder.active_search != wrapper.search_name:
                    self._decoder.active_search = wrapper.search_name

                # TODO Do this in parallel with multiple decoders and Python's multiprocessing package
                hypothesis = self._decoder.batch_process(
                    self._audio_buffers,
                    use_callbacks=False
                )

                if hypothesis:
                    hypothesis = hypothesis.hypstr

                hypothesises[wrapper.search_name] = hypothesis

        # Get the best hypothesis.
        speech = self._get_best_hypothesis(list(hypothesises.values()))

        # Only use wrappers whose search hypothesis is speech.
        wrappers = [w for w in wrappers if hypothesises[w.search_name] == speech]

        processing_states = []
        for wrapper in wrappers:
            # Collect matching rules in a ProcessingState object for each
            # wrapper
            hyp = hypothesises[wrapper.search_name]

            # If called by engine.mimic (not engine.mimic_phrases), we need to match
            # on LinkedRules instead to get a complete match, especially if there's
            # dictation involved.
            if mimicking == "as words":
                processing_states.append(ProcessingState(wrapper, hyp, True))
            else:
                processing_states.append(ProcessingState(wrapper, hyp))

        # Process the best wrapper
        best_state = self._get_best_processing_state(processing_states)
        if best_state:
            best_state.process(notify_only=self._training_session_active)

            # TODO Handle complete sequence rules that can repeat.
            # if best_state.repeatable_complete_sequence_rules:
            #     # Start the repetition timeout timer.
            #     pass
            if best_state.complete_sequence_rules:
                # Now that a complete sequence rule has been processed,
                # clear the engine's in-progress set and reset all wrappers.
                self._clear_in_progress_states_and_reset()

            processing_occurred = True

        # If there are in-progress states, cancel the timer if it is running and
        # start the timer.
        timeout_value = self.config.NEXT_PART_TIMEOUT
        self._cancel_and_free_timeout_timer()
        if self._in_progress_states and timeout_value:
            # Start the timer.
            self._timeout_timer = Timer(
                timeout_value,
                self._on_next_part_timeout
            )
            self._timeout_timer.start()

        # Return whether processing occurred and the final speech hypothesis for
        # post processing.
        return processing_occurred, speech

    def process_buffer(self, buf, sleep_time=0.1):
        """
        Recognise speech from an audio buffer. This method is meant to be called
        in sequence for multiple buffers.

        :param buf: audio buffer
        :param sleep_time: time to sleep after processing an audio buffer. Use 0
            for no sleep time.
        :type sleep_time: float
        :type buf: str
        """
        # Cancel current recognition if it has been requested.
        if self._cancel_recognition_next_time:
            self._decoder.end_utterance()
            self._audio_buffers = []
            self._cancel_recognition_next_time = False

        # Keep a list of buffers for possible reprocessing using different Pocket
        # Sphinx searches later.
        self._audio_buffers.append(buf)

        # Write the buffer to the current .wav file used for acoustic model
        # training. Only do this if in speech, the writer is initialised and
        # recognition is not paused.
        if self._decoder.get_in_speech() and self._training_data_writer and \
                not self.recognition_paused:
            self._training_data_writer.write_to_wave_file(buf)

        # Process audio and wait a few milliseconds.
        self._decoder.process_audio(buf)

        # This improves the performance; we don't need to process as much audio
        # as the device can read, and people don't speak that fast anyway!
        time.sleep(sleep_time)

    def process_wave_file(self, path):
        """
        Recognise speech from a wave file. This method checks that the wave file is
        valid. It raises an error if the file doesn't exist, if it can't be read or
        if the WAV header values do not match those in the engine configuration.

        The wave file must use the same sample width, sample rate and number of
        channels defined in the engine configuration ``PYAUDIO_STREAM_KEYWORD_ARGS``
        attribute.

        If the file is valid, :meth:`process_buffer` is then used to process
        the audio.

        :param path: wave file path
        :raises: IOError | OSError | ValueError
        """
        if not self._decoder:
            self.connect()

        # This method's implementation has been adapted from the PyAudio play wave
        # example: http://people.csail.mit.edu/hubert/pyaudio/#play-wave-example

        # Check that path is a valid file.
        if not os.path.isfile(path):
            raise IOError("'%s' is not a file. Please use a different file path.")

        # Get required audio configuration from the engine config.
        p = PyAudio()
        channels, sample_width, rate, chunk = (
            self.config.PYAUDIO_STREAM_KEYWORD_ARGS["channels"],
            p.get_sample_size(
                self.config.PYAUDIO_STREAM_KEYWORD_ARGS["format"]
            ),
            self.config.PYAUDIO_STREAM_KEYWORD_ARGS["rate"],
            self.config.PYAUDIO_STREAM_KEYWORD_ARGS["frames_per_buffer"]
        )

        # Open the wave file. Use contextlib to make sure that the file is closed
        # whether errors are raised or not.
        with contextlib.closing(wave.open(path, "rb")) as wf:
            # Validate the wave file's header.
            if wf.getnchannels() != channels:
                raise ValueError("WAV file '%s' should use %d channel(s), not %d!"
                                 % (path, channels, wf.getnchannels()))
            elif wf.getsampwidth() != sample_width:
                raise ValueError("WAV file '%s' should use sample width %d, not %d!"
                                 % (path, sample_width, wf.getsampwidth()))
            elif wf.getframerate() != rate:
                raise ValueError("WAV file '%s' should use sample rate %d, not %d!"
                                 % (path, rate, wf.getframerate()))

            # Use process_buffer to process each buffer.
            data = wf.readframes(chunk)
            while data != "":
                self.process_buffer(data, 0)  # no sleep time required
                data = wf.readframes(chunk)

    def post_loader_init(self):
        """
        Do post grammar loader initialisation tasks that can't be done in
        :meth:`connect`.

        This is automatically called by :meth:`recognise_forever`, but **not** by
        :meth:`process_buffer` or :meth:`process_wave_file`.

        This currently only handles the engine's ``START_ASLEEP`` configuration
        option.
        """
        # Start in sleep mode if requested.
        if self.config.START_ASLEEP:
            self.pause_recognition()
            self.log.info("Starting in sleep mode as requested.")

    def recognise_forever(self):
        """
        Start recognising from the default recording device until
        :meth:`disconnect` is called.

        Recognition can be paused and resumed using either the sleep/wake key
        phrases or by calling :meth:`pause_recognition` or
        :meth:`resume_recognition`.

        To configure audio input settings, modify the engine's
        ``PYAUDIO_STREAM_KEYWORD_ARGS`` configuration attribute.
        """
        # Open a PyAudio stream with the specified keyword args and get the number
        # of frames to read per buffer
        p = PyAudio()
        keyword_args = self.config.PYAUDIO_STREAM_KEYWORD_ARGS
        stream = p.open(**keyword_args)
        frames_per_buffer = keyword_args["frames_per_buffer"]

        # Do post grammar loader initialisation.
        self.post_loader_init()

        # Start recognising in a loop
        stream.start_stream()
        self._recognising = True
        self._cancel_recognition_next_time = False
        while self.recognising:
            # Read from the audio device (default number of frames is 2048)
            self.process_buffer(stream.read(frames_per_buffer))

        stream.close()
        self._free_engine_resources()

    def mimic(self, words):
        """ Mimic a recognition of the given *words* """
        if isinstance(words, (list, tuple)):
            words = " ".join(words)

        if self.recognition_paused:
            wake_phrase = self.config.WAKE_PHRASE.strip().lower()
            if words and words.strip().lower() == wake_phrase:
                # Resume if the wake phrase was mimicked.
                self.resume_recognition()

            # Silently return either way.
            return

        # Pretend that Sphinx has started processing speech
        self._speech_start_callback(True)

        # Process the words as if they were spoken
        result = self._hypothesis_callback(words, "as words")
        if not result:
            raise MimicFailure("No matching rule found for words %s." % words)

    def mimic_phrases(self, *phrases):
        """
        Mimic a recognition of the given *phrases*.

        This method accepts variable phrases instead of a list of words to allow
        mimicking of rules using :class:`Dictation` elements.
        """
        if self.recognition_paused:
            words = " ".join(phrases)
            self.mimic(words)  # delegate to mimic()
            return

        # Pretend that Sphinx has started processing speech
        self._speech_start_callback(True)

        # Process phrases as if they were spoken
        for phrase in phrases:
            result = self._hypothesis_callback(phrase, True)
            if not result:
                raise MimicFailure("No matching rule found for words %s."
                                   % phrase)

    def speak(self, text):
        # CMU Sphinx speech recognition engines don't come with text-to-speech
        # functionality. For those who need this, the Jasper project has
        # implementations for popular text-to-speech engines:
        # https://github.com/jasperproject/jasper-client
        raise NotImplementedError("text-to-speech is not implemented for this "
                                  "engine")

    def _get_language(self):
        return self.config.LANGUAGE

    # ---------------------------------------------------------------------
    # Training session methods

    def start_training_session(self):
        """
        Start the training session. This will stop recognition processing
        until either :meth:`end_training_session` is called or the end training
        keyphrase is heard.

        This method is thread safe.
        """
        with self._thread_lock:
            if not self._training_data_writer:
                self._log.warning(
                    "Training data will not be recorded! Please call "
                    "connect() and/or check that engine config "
                    "'TRAINING_DATA_DIR' is set to a valid path.")

            if not self._training_session_active:
                self._log.info("Training session has started. No rule "
                               "actions will be processed. ")
                self._log.info("Say '%s' to end the session."
                               % self.config.END_TRAINING_PHRASE)
                self._training_session_active = True

    def end_training_session(self):
        """
        End the training if one is in progress. This will allow recognition
        processing once again.

        This method is thread safe.
        """
        with self._thread_lock:
            if self._training_session_active:
                self._log.info("Ending training session. Rule actions "
                               "will now be processed normally again.")
                self._training_session_active = False

    # ---------------------------------------------------------------------
    # Recognition loop control methods
    # Stopping recognition loop is done using disconnect()

    @property
    def recognition_paused(self):
        """
        Whether the engine is waiting for the wake phrase to be heard or for
        :meth:`resume_recognition` to be called.

        This property returns False if :meth:`connect` hasn't been called yet.

        :rtype: bool
        """
        if self._decoder:
            return self._decoder.active_search == "_wake_phrase"
        return False

    def pause_recognition(self):
        """
        Pause recognition and wait for :meth:`resume_recognition` to be called or
        for the wake keyphrase to be spoken.

        This method is thread safe.
        """
        with self._thread_lock:
            if not self._decoder:
                return

            # Stop in-progress processing.
            self._clear_in_progress_states_and_reset()

            # Switch to the wake keyphrase search.
            self._decoder.end_utterance()
            self._decoder.active_search = "_wake_phrase"

            # Define temporary callback methods for the decoder.
            def speech_start():
                # Do nothing here for the moment. Another observer notify method
                # could be used if appropriate.
                pass

            def hypothesis(hyp):
                s = hyp.hypstr if hyp else None

                # Resume recognition if s is the wake keyphrase.
                if s and s.strip() == self.config.WAKE_PHRASE.strip():
                    self.resume_recognition()
                else:
                    self.log.debug("Didn't hear %s" % self.config.WAKE_PHRASE)

                # Clear audio buffers
                self._audio_buffers = []

            # Override decoder callback methods.
            self._decoder.speech_start_callback = speech_start
            self._decoder.hypothesis_callback = hypothesis

    def resume_recognition(self):
        """
        Resume listening for grammar rules and key phrases.

        This method is thread safe.
        """
        with self._thread_lock:
            if not self._decoder:
                return

            # Notify observers about recognition resume.
            keyphrase = self.config.WAKE_PHRASE
            words_list = [(word, 0) for word in keyphrase.strip().split()]
            self.observer_manager.notify_recognition(words_list)

            # Restore the callbacks to normal
            def speech_start():
                return self._speech_start_callback(False)

            def hypothesis(hyp):
                # Set speech to the hypothesis string or None if there isn't one
                speech = hyp.hypstr if hyp else None
                return self._hypothesis_callback(speech, False)

            self._decoder.hypothesis_callback = hypothesis
            self._decoder.speech_start_callback = speech_start

            # Switch to an active grammar search / dictation search.
            self._set_current_grammar_search()

    def cancel_recognition(self):
        """
        If a recognition was in progress, cancel it before processing the next
        audio buffer.

        This method is thread safe.
        """
        self._cancel_recognition_next_time = True



#
# This file is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it 
#   under the terms of the GNU Lesser General Public License as published 
#   by the Free Software Foundation, either version 3 of the License, or 
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but 
#   WITHOUT ANY WARRANTY; without even the implied warranty of 
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU 
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public 
#   License along with Dragonfly.  If not, see 
#   <http://www.gnu.org/licenses/>.
#

"""
Context classes
============================================================================

Dragonfly uses context classes to define when grammars and 
rules should be active.  A context is an object with a 
:meth:`Context.matches` method which returns *True* if the 
system is currently within that context, and *False* if it 
is not.

The following context classes are available:

 - :class:`Context` --
   the base class from which all other context classes are derived
 - :class:`AppContext` --
   class which based on the application context, i.e. foreground window
   executable, title, and handle


Logical operations
----------------------------------------------------------------------------

It is possible to modify and combine the behavior of contexts using the
Python's standard logical operators:

:logical AND: ``context1 & context2`` -- *all* contexts must match
:logical OR: ``context1 | context2`` --
   *one or more* of the contexts must match
:logical NOT: ``~context1`` -- *inversion* of when the context matches

For example, to create a context which will match when 
Firefox is in the foreground, but only if Google Reader is 
*not* being viewed::

   firefox_context = AppContext(executable="firefox")
   reader_context = AppContext(executable="firefox", title="Google Reader")
   firefox_but_not_reader_context = firefox_context & ~reader_context


Class reference
----------------------------------------------------------------------------

"""

import copy
import logging


# --------------------------------------------------------------------------
from six import string_types


class Context(object):
    """
        Base class for other context classes.

        This base class implements some basic 
        infrastructure, including what's required for 
        logical operations on context objects.  Derived 
        classes should at least do the following things:

         - During initialization, set *self._str* to some descriptive,
           human readable value.  This attribute is used by the
           ``__str__()`` method.
         - Overload the :meth:`Context.matches` method to implement
           the logic to determine when to be active.

        The *self._log* logger objects should be used in methods of 
        derived classes for logging purposes.  It is a standard logger 
        object from the *logger* module in the Python standard library.

    """

    _log = logging.getLogger("context.match")
    _log_match = _log

    # ----------------------------------------------------------------------
    # Initialization and aggregation methods.

    def __init__(self):
        self._str = ""

    def __str__(self):
        return "%s(%s)" % (self.__class__.__name__, self._str)

    def copy(self):
        return copy.deepcopy(self)

    # ----------------------------------------------------------------------
    # Logical operations.

    def __and__(self, other):
        return LogicAndContext(self, other)

    def __or__(self, other):
        return LogicOrContext(self, other)

    def __invert__(self):
        return LogicNotContext(self)

    # ----------------------------------------------------------------------
    # Matching methods.

    def matches(self, executable, title, handle):
        """
            Indicate whether the system is currently within this context.

            Arguments:
             - *executable* (*str*) --
               path name to the executable of the foreground application
             - *title* (*str*) -- title of the foreground window
             - *handle* (*int*) -- window handle to the foreground window

            The default implementation of this method simply returns *True*.

            .. note::

               This is generally the method which developers should
               overload to give derived context classes custom
               functionality.

        """
        return True


# --------------------------------------------------------------------------
# Wrapper contexts for combining contexts in logical structures.

class LogicAndContext(Context):

    def __init__(self, *children):
        self._children = children
        self._str = ", ".join(str(child) for child in children)

    def matches(self, executable, title, handle):
        for child in self._children:
            if not child.matches(executable, title, handle):
                return False
        return True


class LogicOrContext(Context):

    def __init__(self, *children):
        self._children = children
        self._str = ", ".join(str(child) for child in children)

    def matches(self, executable, title, handle):
        for child in self._children:
            if child.matches(executable, title, handle):
                return True
        return False


class LogicNotContext(Context):

    def __init__(self, child):
        self._child = child
        self._str = str(child)

    def matches(self, executable, title, handle):
        return not self._child.matches(executable, title, handle)


# --------------------------------------------------------------------------

class AppContext(Context):
    """
        Context class using foreground application details.

        This class determines whether the foreground window meets
        certain requirements.  Which requirements must be met for this
        context to match are determined by the constructor arguments.

        Constructor arguments:
         - *executable* (*str*) --
           (part of) the path name of the foreground application's
           executable; case insensitive
         - *title* (*str*) --
           (part of) the title of the foreground window; case insensitive

    """

    # ----------------------------------------------------------------------
    # Initialization methods.

    def __init__(self, executable=None, title=None, exclude=False):
        Context.__init__(self)

        # Allow Unicode or strings to be used for executables and titles.
        if isinstance(executable, string_types):
            self._executable = executable.lower()
        elif executable is None:
            self._executable = None
        else:
            raise TypeError("executable argument must be a string or None;"
                            " received %r" % executable)

        if isinstance(title, string_types):
            self._title = title.lower()
        elif title is None:
            self._title = None
        else:
            raise TypeError("title argument must be a string or None;"
                            " received %r" % title)

        self._exclude = bool(exclude)

        self._str = "%s, %s, %s" % (self._executable, self._title,
                                    self._exclude)

    # ----------------------------------------------------------------------
    # Matching methods.

    def matches(self, executable, title, handle):
        executable = executable.lower()
        title = title.lower()

        if self._executable:
            found = (executable.find(self._executable) != -1)
            if self._exclude == found:
                self._log_match.debug("%s: No match, executable doesn't"
                                      " match." % self)
                return False
        if self._title:
            found = (title.find(self._title) != -1)
            if self._exclude == found:
                self._log_match.debug("%s: No match, title doesn't match."
                                      % self)
                return False

        if self._log_match:
            self._log_match.debug("%s: Match." % self)
        return True



"""
Command-module loader for CMU Pocket Sphinx.

This script is based on 'dfly-loader-wsr.py' written by Christo Butcher and
has been adapted to work with the Sphinx engine instead.

This script can be used to look for Dragonfly command-modules for use with
the CMU Pocket Sphinx engine. It scans the directory it's in and loads any
``_*.py`` it finds.
"""


import os.path
import logging


from dragonfly import RecognitionObserver, EngineError
from dragonfly.engines import get_engine
from dragonfly.engines.backend_sphinx.engine import SphinxEngine
from dragonfly.loader import CommandModuleDirectory

# --------------------------------------------------------------------------
# Set up basic logging.

logging.basicConfig(level=logging.INFO)
logging.getLogger("compound.parse").setLevel(logging.INFO)


# --------------------------------------------------------------------------
# Simple recognition observer class.

class Observer(RecognitionObserver):
    def __init__(self, engine):
        self.engine = engine
        super(Observer, self).__init__()

    def on_begin(self):
        if self.engine.recognising_dictation:
            print("Speech started. Processing as dictation.")
        else:
            print("Speech started using grammar search.")

    @staticmethod
    def _get_words(words_list):
        # Get just the words from the tuple list
        return " ".join([word for word, _ in words_list])

    def on_recognition(self, words_list):
        print(self._get_words(words_list))

    def on_failure(self):
        print("Sorry, what was that?")

    def on_next_rule_part(self, words_list):
        print("Current words: %s" % self._get_words(words_list))
        print("Awaiting next rule part...")


# --------------------------------------------------------------------------
# Main event driving loop.

def main():
    logging.basicConfig(level=logging.INFO)

    try:
        path = os.path.dirname(__file__)
    except NameError:
        # The "__file__" name is not always available, for example
        # when this module is run from PythonWin.  In this case we
        # simply use the current working directory.
        path = os.getcwd()
        __file__ = os.path.join(path, "sphinx_module_loader.py")

    engine = get_engine("sphinx")
    assert isinstance(engine, SphinxEngine)

    # Try to import the local engine configuration object first. If there isn't one,
    # use the default engine configuration.
    log = logging.getLogger("config")
    try:
        import config
        engine.config = config
        log.info("Using local engine configuration module 'config.py'")
    except ImportError:
        pass
    except EngineError as e:
        # Log EngineErrors caught when setting the configuration.
        log.warning(e)
        log.warning("Falling back to using the default engine configuration "
                    "instead of 'config.py'")

    # Call connect() now that the engine configuration is set.
    engine.connect()

    # Register a recognition observer
    observer = Observer(engine)
    observer.register()

    directory = CommandModuleDirectory(path, excludes=[__file__])
    directory.load()

    # TODO Have a simple GUI for pausing, resuming, cancelling and stopping recognition, etc
    # TODO Change script to import all modules before loading the grammars into Pocket Sphinx

    # Start the engine's main recognition loop
    try:
        engine.recognise_forever()
    except KeyboardInterrupt:
        pass


if __name__ == "__main__":
    main()



import sys
sys.path.pop(0)

import nose
from dragonfly.engines.backend_natlink import test_dictation
from dragonfly import *

if __name__ == "__main__":
    engine = get_engine()
    engine.connect()
    try:
        loader = nose.loader.TestLoader()
        suite = loader.loadTestsFromTestCase(test_dictation.EnglishNatlinkDictationTestCase)
        nose.core.TestProgram(suite=suite)
    finally:
        engine.disconnect()



# -*- encoding: utf-8 -*-
#
# This file is part of Dragonfly.
# (c) Copyright 2018 by Dane Finlay
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it 
#   under the terms of the GNU Lesser General Public License as published 
#   by the Free Software Foundation, either version 3 of the License, or 
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but 
#   WITHOUT ANY WARRANTY; without even the implied warranty of 
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU 
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public 
#   License along with Dragonfly.  If not, see 
#   <http://www.gnu.org/licenses/>.
#

import unittest

from dragonfly.engines import EngineBase
from dragonfly import (Literal, Dictation, Sequence, Repetition,
                       CompoundRule, MimicFailure, get_engine, List,
                       DictList, ListRef, DictListRef, RecognitionHistory,
                       Rule)
from dragonfly.engines.backend_text.dictation import TextDictationContainer
from dragonfly.test import (ElementTester, RecognitionFailure, RuleTestCase,
                            TestContext, RuleTestGrammar)


# --------------------------------------------------------------------------

class TestEngineText(unittest.TestCase):

    def setUp(self):
        engine = get_engine("text")
        assert isinstance(engine, EngineBase)
        assert engine.name == "text"
        engine.connect()
        self.engine = engine

    def tearDown(self):
        self.engine.disconnect()

    def test_literal(self):
        """ Verify that the text engine is usable. """
        tester = ElementTester(Literal("hello world"))
        results = tester.recognize("hello world")
        assert results == "hello world"

        # Check that recognition failure is possible.
        results = tester.recognize("goodbye")
        assert results is RecognitionFailure

    def test_dictation(self):
        """ Verify that the text engine can mimic dictation. """
        tester = ElementTester(Dictation("text"))
        results = tester.recognize("HELLO WORLD")
        assert isinstance(results, TextDictationContainer)
        assert results.format() == "hello world"

        # Check that an empty string results in recognition failure.
        results = tester.recognize("")
        assert results is RecognitionFailure

    def test_mixed_dictation(self):
        """ Verify that the text engine can mimic rules with Dictation and
            Literal elements. """
        seq = Sequence([Literal("hello"), Dictation("text")])
        tester = ElementTester(seq)
        results = tester.recognize("hello WORLD")
        assert results[0] == "hello"
        assert isinstance(results[1], TextDictationContainer)
        assert results[1].format() == "world"

        # Check that strings not starting with "hello" result in recognition
        # failure.
        results = tester.recognize("goodbye world")
        assert results is RecognitionFailure

    def test_unicode_literals(self):
        """ Verify that the text engine can mimic literals using non-ascii
            characters. """
        tester = ElementTester(Literal(u"Привет, как дела?"))
        results = tester.recognize(u"Привет, как дела?")
        assert results == u"Привет, как дела?"

        # Check that recognition failure is possible.
        results = tester.recognize(u"Привет")
        assert results is RecognitionFailure
        results = tester.recognize(u"до свидания")
        assert results is RecognitionFailure

    def test_repetition(self):
        """ Verify that the text engine can mimic a rule using a Repetition
            element. """
        rep = Repetition(Sequence([Literal("hello"), Dictation("text")]),
                         min=1, max=16)
        tester = ElementTester(rep)

        # Test with one repetition.
        results = tester.recognize("hello WORLD")
        assert len(results) == 1
        assert results[0][0] == "hello"
        assert isinstance(results[0][1], TextDictationContainer)
        assert results[0][1].format() == "world"

        # Test with two repetitions.
        results = tester.recognize("hello WORLD hello TESTING")
        assert len(results) == 2
        assert results[0][0] == "hello"
        assert isinstance(results[0][1], TextDictationContainer)
        assert results[0][1].format() == "world"
        assert results[1][0] == "hello"
        assert isinstance(results[1][1], TextDictationContainer)
        assert results[1][1].format() == "testing"

        # Check that incomplete repetitions result in recognition
        # failure.
        results = tester.recognize("hello")
        assert results is RecognitionFailure
        results = tester.recognize("hello world")
        assert results is RecognitionFailure
        results = tester.recognize("hello WORLD hello")
        assert results is RecognitionFailure


class TestRules(RuleTestCase):
    def test_multiple_rules(self):
        """ Verify that the text engine successfully mimics each rule in a
            grammar with multiple rules. """
        self.add_rule(CompoundRule(name="r1", spec="hello"))
        self.add_rule(CompoundRule(name="r2", spec="see you"))
        assert self.recognize_node("hello").words() == ["hello"]
        assert self.recognize_node("see you").words() == ["see", "you"]

    def test_list(self):
        """ Verify that the text engine can mimic a rule using a ListRef
            element. """
        lst = List("fruit")
        lst.append("apple")
        self.grammar.add_list(lst)
        self.add_rule(Rule("fav_fruit", ListRef("fruit_ref", lst),
                           exported=True))

        # Test that the list item "apple" can be recognized.
        assert self.recognize_node("apple").words() == ["apple"]

        # Update the list and try again.
        lst.append("banana")

        # If recognition fails, then the list wasn't updated correctly.
        assert self.recognize_node("banana").words() == ["banana"]

    def test_dict_list(self):
        """ Verify that the text engine can mimic a rule using a DictListRef
            element. """
        lst = DictList("fruit")
        lst["mango"] = False
        self.grammar.add_list(lst)
        self.add_rule(Rule("fav_fruit", DictListRef("fruit_ref", lst),
                           exported=True))

        # Test that the list item "mango" can be recognized.
        assert self.recognize_node("mango").words() == ["mango"]

        # Update the list and try again.
        lst["mandarin"] = True

        # If recognition fails, then the list wasn't updated correctly.
        assert self.recognize_node("mandarin").words() == ["mandarin"]

    def test_rule_context(self):
        """ Verify that the text engine works correctly with rule
            contexts."""
        context = TestContext(True)
        self.add_rule(CompoundRule(name="r1", spec="test context",
                                   context=context))

        # Test that the rule matches when in-context.
        results = self.recognize_node("test context").words()
        assert results == ["test", "context"]

        # Go out of context and test again.
        # Use the engine's mimic method because recognize_node won't return
        # RecognitionFailure like ElementTester.recognize does.
        context.active = False
        self.assertRaises(MimicFailure, self.engine.mimic, "test context")

        # Test again after going back into context.
        context.active = True
        results = self.recognize_node("test context").words()
        assert results == ["test", "context"]

    def test_grammar_context(self):
        """ Verify that the text engine works correctly with grammar
            contexts."""
        # Recreate the RuleTestGrammar using a context and add a rule.
        context = TestContext(True)
        self.grammar = RuleTestGrammar(context=context)
        self.add_rule(CompoundRule(name="r1", spec="test context"))

        # Test that the rule matches when in-context.
        results = self.recognize_node("test context").words()
        assert results == ["test", "context"]

        # Go out of context and test again.
        context.active = False
        self.assertRaises(MimicFailure, self.engine.mimic, "test context")

        # Test again after going back into context.
        context.active = True
        results = self.recognize_node("test context").words()
        assert results == ["test", "context"]

    def test_exclusive_grammars(self):
        """ Verify that the text engine supports exclusive grammars. """
        # Set up two grammars to test with.
        grammar1 = self.grammar
        grammar1.add_rule(CompoundRule(spec="grammar one"))
        grammar2 = RuleTestGrammar(name="Grammar2")
        grammar2.add_rule(CompoundRule(spec="grammar two"))
        grammar1.load()
        grammar2.load()

        # Set grammar1 as exclusive and make some assertions.
        grammar1.set_exclusiveness(True)
        results = grammar1.recognize_node("grammar one").words()
        assert results == ["grammar", "one"]
        self.assertRaises(MimicFailure, self.engine.mimic, "grammar two")

        # Set grammar1 as no longer exclusive and make some assertions.
        grammar1.set_exclusiveness(False)
        results = grammar1.recognize_node("grammar one").words()
        assert results == ["grammar", "one"]
        results = grammar2.recognize_node("grammar two").words()
        assert results == ["grammar", "two"]

    def test_recognition_observers(self):
        """ Verify that the text engine notifies recognition observers
            correctly. """
        class TestRecognitionHistory(RecognitionHistory):
            def __init__(self, length=10):
                super(TestRecognitionHistory, self).__init__(length)
                self.begins = 0
                self.failures = 0

            def on_begin(self):
                super(TestRecognitionHistory, self).on_begin()
                self.begins += 1

            def on_failure(self):
                self.failures += 1

        history = TestRecognitionHistory()
        history.register()
        self.add_rule(CompoundRule(name="r1", spec="hello"))
        self.add_rule(CompoundRule(name="r2", spec="see you"))

        # Mimic twice and check that history received the words.
        assert self.recognize_node("hello").words() == ["hello"]
        assert self.recognize_node("see you").words() == ["see", "you"]
        assert history == [["hello"], ["see", "you"]]

        # Check that on_begin() was called twice.
        assert history.begins == 2

        # Check that on_begin() and on_failure() get called for failed
        # recognitions.
        self.assertRaises(MimicFailure, self.engine.mimic, "test failure")
        assert history.begins == 3
        assert history.failures == 1



#
# This file is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it 
#   under the terms of the GNU Lesser General Public License as published 
#   by the Free Software Foundation, either version 3 of the License, or 
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but 
#   WITHOUT ANY WARRANTY; without even the implied warranty of 
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU 
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public 
#   License along with Dragonfly.  If not, see 
#   <http://www.gnu.org/licenses/>.
#

"""
SAPI 5 engine classes
============================================================================

"""


#---------------------------------------------------------------------------

import time
from six import string_types, integer_types

import win32con
from ctypes import *

from win32com.client           import Dispatch, getevents, constants
from win32com.client.gencache  import EnsureDispatch
from pywintypes                import com_error

from ..base                    import EngineBase, EngineError
from .compiler                 import Sapi5Compiler
from .dictation                import Sapi5DictationContainer
from .recobs                   import Sapi5RecObsManager
#from .timer                    import NatlinkTimerManager
from ...grammar.state          import State
from ...windows.window         import Window


#===========================================================================

class POINT(Structure):
    _fields_ = [('x', c_long),
                ('y', c_long)]

class MSG(Structure):
    _fields_ = [('hwnd', c_int),
                ('message', c_uint),
                ('wParam', c_int),
                ('lParam', c_int),
                ('time', c_int),
                ('pt', POINT)]


#===========================================================================

class Sapi5SharedEngine(EngineBase):
    """ Speech recognition engine back-end for SAPI 5 shared recognizer. """

    _name = "sapi5shared"
    _recognizer_dispatch_name = "SAPI.SpSharedRecognizer"
    DictationContainer = Sapi5DictationContainer

    #-----------------------------------------------------------------------

    def __init__(self):
        EngineBase.__init__(self)

        EnsureDispatch(self._recognizer_dispatch_name)
        EnsureDispatch("SAPI.SpVoice")
        self._recognizer  = None
        self._speaker     = None
        self._compiler    = None

        self._recognition_observer_manager = Sapi5RecObsManager(self)
#        self._timer_manager = NatlinkTimerManager(0.02, self)

    def connect(self):
        """ Connect to back-end SR engine. """
        self._recognizer  = Dispatch(self._recognizer_dispatch_name)
        self._speaker     = Dispatch("SAPI.SpVoice")
        self._compiler    = Sapi5Compiler()

    def disconnect(self):
        """ Disconnect from back-end SR engine. """
        self._recognizer  = None
        self._speaker     = None
        self._compiler    = None

    #-----------------------------------------------------------------------
    # Methods for working with grammars.

    def _load_grammar(self, grammar):
        """ Load the given *grammar*. """
        self._log.debug("Loading grammar %s." % grammar.name)
        if not self._recognizer:
            self.connect()

        grammar.engine = self

        # Dependency checking.
        memo = []
        for r in grammar._rules:
            for d in r.dependencies(memo):
                grammar.add_dependency(d)

        # Create recognition context, compile grammar, and create
        #  the grammar wrapper object for managing this grammar.
        context = self._recognizer.CreateRecoContext()
        handle = self._compiler.compile_grammar(grammar, context)
        wrapper = GrammarWrapper(grammar, handle, context, self)

        handle.State = constants.SGSEnabled
        for rule in grammar.rules:
            handle.CmdSetRuleState(rule.name, constants.SGDSActive)
#        self.activate_grammar(grammar)
#        for l in grammar.lists:
#            l._update()
        handle.CmdSetRuleState("_FakeRule", constants.SGDSActive)

        return wrapper

    def _unload_grammar(self, grammar, wrapper):
        """ Unload the given *grammar*. """
        try:
            wrapper.handle.State = constants.SGSDisabled
        except Exception as e:
            self._log.exception("Failed to unload grammar %s: %s."
                                % (grammar, e))

    def activate_grammar(self, grammar):
        """ Activate the given *grammar*. """
        self._log.debug("Activating grammar %s." % grammar.name)
        grammar_handle = self._get_grammar_wrapper(grammar).handle
        grammar_handle.State = constants.SGSEnabled

    def deactivate_grammar(self, grammar):
        """ Deactivate the given *grammar*. """
        self._log.debug("Deactivating grammar %s." % grammar.name)
        grammar_handle = self._get_grammar_wrapper(grammar).handle
        grammar_handle.State = constants.SGSDisabled

    def activate_rule(self, rule, grammar):
        """ Activate the given *rule*. """
        self._log.debug("Activating rule %s in grammar %s."
                        % (rule.name, grammar.name))
        grammar_handle = self._get_grammar_wrapper(grammar).handle
        grammar_handle.CmdSetRuleState(rule.name, constants.SGDSActive)

    def deactivate_rule(self, rule, grammar):
        """ Deactivate the given *rule*. """
        self._log.debug("Deactivating rule %s in grammar %s."
                        % (rule.name, grammar.name))
        grammar_handle = self._get_grammar_wrapper(grammar).handle
        grammar_handle.CmdSetRuleState(rule.name, constants.SGDSInactive)

    def update_list(self, lst, grammar):
        grammar_handle = self._get_grammar_wrapper(grammar).handle
        list_rule_name = "__list_%s" % lst.name
        rule_handle = grammar_handle.Rules.FindRule(list_rule_name)

        rule_handle.Clear()
        src_state = rule_handle.InitialState
        dst_state = None
        for item in lst.get_list_items():
            src_state.AddWordTransition(dst_state, item)

        grammar_handle.Rules.Commit()

    def set_exclusiveness(self, grammar, exclusive):
        self._log.debug("Setting exclusiveness of grammar %s to %s."
                        % (grammar.name, exclusive))
        grammar_handle = self._get_grammar_wrapper(grammar).handle
        grammar_handle.State = constants.SGSExclusive
#        grammar_handle.SetGrammarState(constants.SPGS_EXCLUSIVE)


    #-----------------------------------------------------------------------
    # Miscellaneous methods.

    def mimic(self, words):
        """ Mimic a recognition of the given *words*. """
        if isinstance(words, string_types):
            phrase = words
        else:
            phrase = " ".join(words)
        self._recognizer.EmulateRecognition(phrase)

    def speak(self, text):
        """ Speak the given *text* using text-to-speech. """
        self._speaker.Speak(text)

    def _get_language(self):
        return "en"

    def wait_for_recognition(self, timeout=None):
        NULL = c_int(win32con.NULL)
        if timeout != None:
            begin_time = time.time()
            timed_out = False
            windll.user32.SetTimer(NULL, NULL, int(timeout * 1000), NULL)
    
        message = MSG()
        message_pointer = pointer(message)

        while (not timeout) or (time.time() - begin_time < timeout):
            self._log.error("loop")
            if windll.user32.GetMessageW(message_pointer, NULL, 0, 0) == 0:
                msg = str(WinError())
                self._log.error("GetMessageW() failed: %s" % msg)
                raise EngineError("GetMessageW() failed: %s" % msg)

            if message.message == win32con.WM_TIMER:
                self._log.error("loop, timeout")
                # A timer message means this loop has timed out.
                timed_out = True
                break
            else:
                self._log.error("loop, dispatch")
                # Process other messages as normal.
                windll.user32.TranslateMessage(message_pointer)
                windll.user32.DispatchMessageW(message_pointer)

        return not timed_out


#---------------------------------------------------------------------------
# Make the shared engine available as Sapi5Engine, for backwards
#  compatibility.

Sapi5Engine = Sapi5SharedEngine


#===========================================================================

class Sapi5InProcEngine(Sapi5SharedEngine):
    """
        Speech recognition engine back-end for SAPI 5 in process
        recognizer.

    """

    _name = "sapi5inproc"
    _recognizer_dispatch_name = "SAPI.SpInProcRecognizer"

    def connect(self, audio_source=0):
        """
            Connect to the speech recognition backend.

            The audio source to use for speech recognition can be
            specified using the *audio_source* argument. If it is not
            given, it defaults to the first audio source found.

        """

        Sapi5SharedEngine.connect(self)
        self.select_audio_source(audio_source)

    def get_audio_sources(self):
        """
            Get the available audio sources.

            This method returns a list of audio sources, each represented
            by a 3-element tuple: the index, the description, and the COM
            handle for the audio source.

        """

        available_sources = self._recognizer.GetAudioInputs()
        audio_sources_list = []
        for index, item in enumerate(collection_iter(available_sources)):
            audio_sources_list.append((index, item.GetDescription(), item))
        return audio_sources_list

    def select_audio_source(self, audio_source):
        """
            Configure the speech recognition engine to use the given
            audio source.

            The audio source may be specified as follows:
             - As an *int* specifying the index of the audio source to use
             - As a *str* containing the description of the audio source
               to use, or a substring thereof

            This class' method *get_audio_sources()* can be used to
            retrieve the available sources together with their indices
            and descriptions.

        """

        available_sources = self._recognizer.GetAudioInputs()

        if isinstance(audio_source, integer_types):
            # Parameter is the index of the source to use.
            if 0 <= audio_source < available_sources.Count:
                selected_source = available_sources.Item(audio_source)
            else:
                raise EngineError("Invalid audio source index: %r"
                                  " (%s sources available, so index must be"
                                  " in range 0 to %s)"
                                  % (audio_source, available_sources.Count,
                                     available_sources.Count - 1))

        elif isinstance(audio_source, string_types):
            for item in collection_iter(available_sources):
                if audio_source in item.GetDescription():
                    selected_source = item
                    break
            else:
                raise EngineError("Audio source not found: %r"
                                  % (audio_source))

        else:
            raise EngineError("Invalid audio source qualifier: %r"
                              % (audio_source))

        self._log.info("Selecting audio source: %r"
                       % (selected_source.GetDescription(),))
        self._recognizer.AudioInput = selected_source


#---------------------------------------------------------------------------
# Utility generator function for iterating over COM collections.

def collection_iter(collection):
    if not collection:
        return
    for index in range(0, collection.Count):
        yield collection.Item(index)


#---------------------------------------------------------------------------

class GrammarWrapper(object):

    def __init__(self, grammar, handle, context, engine):
        self.grammar = grammar
        self.handle = handle
        self.engine = engine
        self.context = context

        # Register callback functions which will handle recognizer events.
        base = getevents("SAPI.SpSharedRecoContext")
        class ContextEvents(base): pass
        c = ContextEvents(context)
        c.OnPhraseStart = self.phrase_start_callback
        c.OnRecognition = self.recognition_callback
        if hasattr(grammar, "process_recognition_other"):
            c.OnRecognitionForOtherContext = self.recognition_other_callback
        if hasattr(grammar, "process_recognition_failure"):
            c.OnFalseRecognition = self.recognition_failure_callback

    def phrase_start_callback(self, stream_number, stream_position):
        window = Window.get_foreground()
        self.grammar.process_begin(window.executable, window.title,
                                   window.handle)

    def recognition_callback(self, StreamNumber, StreamPosition, RecognitionType, Result):
        try:
            newResult = Dispatch(Result)
            phrase_info = newResult.PhraseInfo
            rule_name = phrase_info.Rule.Name

            #---------------------------------------------------------------
            # Build a list of rule names for each element.

            # First populate it with the top level rule name.
            element = phrase_info.Rule
            name = element.Name
            start = element.FirstElement
            count = element.NumberOfElements
            rule_names = [name] * count

            # Walk the tree of child rules and put their names in the list.
            stack = [collection_iter(phrase_info.Rule.Children)]
            while stack:
                try: element = next(stack[-1])
                except StopIteration: stack.pop(); continue
                name = element.Name
                start = element.FirstElement
                count = element.NumberOfElements
                rule_names[start:start + count] = [name] * count
                if element.Children:
                    stack.append(collection_iter(element.Children))

            #---------------------------------------------------------------
            # Prepare the words and rule names for the element parsers.

            replacements = [False] * len(rule_names)
            if phrase_info.Replacements:
                for replacement in collection_iter(phrase_info.Replacements):
                    begin = replacement.FirstElement
                    end = begin + replacement.NumberOfElements
                    replacements[begin] = replacement.Text
                    for index in range(begin + 1, end):
                        replacements[index] = True

            results = []
            rule_set = list(set(rule_names))
            elements = phrase_info.Elements
            for index in range(len(rule_names)):
                element = elements.Item(index)
                rule_id = rule_set.index(rule_names[index])
                replacement = replacements[index]
                info = [element.LexicalForm, rule_id,
                        element.DisplayText, element.DisplayAttributes,
                        replacement]
                results.append(info)

            #---------------------------------------------------------------
            # Attempt to parse the recognition.

            func = getattr(self.grammar, "process_recognition", None)
            if func:
                words = [r[2] for r in results]
                if not func(words):
                    return

            s = State(results, rule_set, self.engine)
            for r in self.grammar._rules:
                if r.name != rule_name:
                    continue
                s.initialize_decoding()
                for result in r.decode(s):
                    if s.finished():
                        root = s.build_parse_tree()
                        r.process_recognition(root)
                        return

        except Exception as e:
            Sapi5Engine._log.error("Grammar %s: exception: %s"
                                   % (self.grammar._name, e), exc_info=True)

        #-------------------------------------------------------------------
        # If this point is reached, then the recognition was not
        #  processed successfully..

        self.engine._log.error("Grammar %s: failed to decode"
                               " recognition %r."
                               % (self.grammar._name,
                                  [r[0] for r in results]))

    def recognition_other_callback(self, StreamNumber, StreamPosition):
            func = getattr(self.grammar, "process_recognition_other", None)
            if func:
                # Note that SAPI 5.3 doesn't offer access to the actual
                #  recognition contents during a
                #  OnRecognitionForOtherContext event.
                func(words=False)
            return

    def recognition_failure_callback(self, StreamNumber, StreamPosition, Result):
            func = getattr(self.grammar, "process_recognition_failure", None)
            if func:
                func()
            return



# -*- encoding: windows-1252 -*-
#
# This file is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it 
#   under the terms of the GNU Lesser General Public License as published 
#   by the Free Software Foundation, either version 3 of the License, or 
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but 
#   WITHOUT ANY WARRANTY; without even the implied warranty of 
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU 
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public 
#   License along with Dragonfly.  If not, see 
#   <http://www.gnu.org/licenses/>.
#


import unittest
from six import string_types, text_type

from dragonfly.engines.base.dictation   import DictationContainerBase
from dragonfly.grammar.elements         import Compound, Dictation
from dragonfly.test.infrastructure      import RecognitionFailure
from dragonfly.test.element_testcase    import ElementTestCase
from dragonfly.test.element_tester      import ElementTester


#===========================================================================

class TestNonAsciiDictation(unittest.TestCase):

    def test_dictation_non_ascii(self):
        """ Test handling of non-ASCII characters in dictation. """

        def value_func(node, extras):
            return extras["text"]
        element = Compound("test <text>",
                           extras=[Dictation(name="text")],
                           value_func=value_func)
        tester = ElementTester(element)

        words = [u"test", u"touché"]
        dictation = tester.recognize(words)

        # Verify recognition returned dictation result.
        if not isinstance(dictation, DictationContainerBase):
            message = (u"Expected recognition result to be a dictation"
                       u" container, but received %r"
                       % (repr(dictation).decode("windows-1252"),))
            self.fail(message.encode("windows-1252"))

        # Verifying dictation converts/encode successfully.
        self.assertEqual(str(dictation), "touché")
        self.assertEqual(text_type(dictation), u"touché")
        self.assertTrue(isinstance(repr(dictation), string_types))


class NonAsciiUnicodeDictationTestCase(ElementTestCase):
    """ Verify handling of non-ASCII characters in unicode dictation. """

    def _build_element(self):
        def value_func(node, extras):
            return text_type(extras["text"])
        return Compound("test <text>",
                        extras=[Dictation(name="text")],
                        value_func=value_func)

    input_output = [
                    (u"test dictation",     u"dictation"),
                    (u"test touché",        u"touché"),
                    (u"test jalapeño",      u"jalapeño"),
                   ]


class NonAsciiStrDictationTestCase(ElementTestCase):
    """ Verify handling of non-ASCII characters in str dictation. """

    def _build_element(self):
        def value_func(node, extras):
            return str(extras["text"])
        return Compound("test <text>",
                        extras=[Dictation(name="text")],
                        value_func=value_func)

    input_output = [
                    ("test dictation",     "dictation"),
                    ("test touché",        "touché"),
                    ("test jalapeño",      "jalapeño"),
                   ]


#===========================================================================

if __name__ == "__main__":
    unittest.main()



#
# This file is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it 
#   under the terms of the GNU Lesser General Public License as published 
#   by the Free Software Foundation, either version 3 of the License, or 
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but 
#   WITHOUT ANY WARRANTY; without even the implied warranty of 
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU 
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public 
#   License along with Dragonfly.  If not, see 
#   <http://www.gnu.org/licenses/>.
#

"""
Dictation container base class
============================================================================

This class is used to store the recognized results of dictation elements 
within voice-commands.  It offers access to both the raw spoken-form words 
and be formatted written-form text.

The formatted text can be retrieved using
:meth:`~DictationContainerBase.format` or simply by  calling ``str(...)``
on a dictation container object. A tuple of the raw  spoken words can be
retrieved using :attr:`~DictationContainerBase.words`.

"""


#---------------------------------------------------------------------------
# Dictation base class -- base class for SR engine-specific containers
#  of dictated words.

class DictationContainerBase(object):
    """
        Container class for dictated words as recognized by the
        :class:`Dictation` element.

        This base class implements the general functionality of dictation 
        container classes.  Each supported engine should have a derived 
        dictation container class which performs the actual engine-
        specific formatting of dictated text.

    """

    def __init__(self, words):
        """
            A dictation container is created by passing it a sequence
            of words as recognized by the backend SR engine.
            Each word must be a Unicode string.

            :param words: A sequence of Unicode strings.
            :type words: sequence-of-unicode

        """
        self._words = tuple(words)
        self._formatted = None

    def __str__(self):
        if self._formatted is None:
            self._formatted = self.format()
        return self._formatted

    def __unicode__(self):
        if self._formatted is None:
            self._formatted = self.format()
        return self._formatted

    def __repr__(self):
        message = u"%s(%s)" % (self.__class__.__name__,
                               u", ".join(self._words))
        return message.encode("windows-1252")

    @property
    def words(self):
        """ Sequence of the words forming this dictation. """
        return self._words

    def format(self):
        """ Format and return this dictation as a Unicode object. """
        return u" ".join(self._words)



#
# This file is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it 
#   under the terms of the GNU Lesser General Public License as published 
#   by the Free Software Foundation, either version 3 of the License, or 
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but 
#   WITHOUT ANY WARRANTY; without even the implied warranty of 
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU 
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public 
#   License along with Dragonfly.  If not, see 
#   <http://www.gnu.org/licenses/>.
#

"""
Tools for testing element classes
============================================================================

"""

import logging
from six import string_types

from dragonfly              import *
from ..test                 import TestError, RecognitionFailure
from ..test.infrastructure  import Unique


#===========================================================================

class ElementTester(Grammar):

    _log = logging.getLogger("test.element")
    _NotSet = Unique("NoRecognition")

    class _ElementTestRule(Rule):
        exported = True
        def process_recognition(self, node):
            self.grammar._process_recognition(node)

    #-----------------------------------------------------------------------

    def __init__(self, element, engine=None):
        Grammar.__init__(self, self.__class__.__name__, engine=engine)
        rule = self._ElementTestRule("rule", element)
        self.add_rule(rule)

    def recognize(self, words):
        if isinstance(words, string_types):
            words = words.split()

        if not self.loaded:
            self._log.debug("Loading ElementTester instance.")
            self.load()
            unload_after_recognition = True
        else:
            unload_after_recognition = False

        self._recognized_value = self._NotSet
        try:
            # Make this grammar exclusive; this *probably* avoids other
            #  grammars from being active and receiving the mimicked
            #  recognition.
            self.set_exclusiveness(True)

            # Mimic recognition.
            try:
                mimic_method = self._mimic_methods[self.engine.name]
                mimic_method(self, words)
            except MimicFailure as e:
                self._recognized_value = RecognitionFailure
            except Exception as e:
                self._log.exception("Exception within recognition: %s" % (e,))
                raise

        except Exception as e:
            self._log.exception("Exception during recognition: %s" % (e,))
            raise
        finally:
            if unload_after_recognition:
                try:
                    self.unload()
                except Exception as e:
                    raise TestError("Failed to unload grammar: %s" % e)

        # If recognition was successful but this grammar did not
        #  receive the recognition callback, then apparently some other
        #  grammar hijacked it; raise a TestError to signal this
        #  undesired situation.
        if self._recognized_value == self._NotSet:
            self._log.error(u"Recognition hijacked. (Words: %s)" % (words,))
            raise TestError(u"Recognition hijacked. (Words: %s)" % (words,))

        # Return the value of the element after recognition.
        return self._recognized_value

    def _process_recognition(self, node):
        element_node = node.children[0]
        self._recognized_value = element_node.value()


    #-----------------------------------------------------------------------
    # Engine-specific logic.

    _mimic_methods = {}

    def _mimic_words(self, words):
        self.engine.mimic(words)

    _mimic_methods["natlink"] = _mimic_words
    _mimic_methods["sphinx"] = _mimic_words
    _mimic_methods["text"] = _mimic_words

    def _mimic_sapi5(self, words):
        import time
        import win32con
        from ctypes import (windll, pointer, WinError, Structure,
                            c_int, c_uint, c_long)

        class POINT(Structure):
            _fields_ = [('x', c_long),
                        ('y', c_long)]

        class MSG(Structure):
            _fields_ = [('hwnd', c_int),
                        ('message', c_uint),
                        ('wParam', c_int),
                        ('lParam', c_int),
                        ('time', c_int),
                        ('pt', POINT)]

        class Obs(RecognitionObserver):
            _log = logging.getLogger("SAPI5 RecObs")
            status = "none"
            def on_recognition(self, words):
                self._log.debug("SAPI5 RecObs on_recognition(): %r" % (words,))
                self.status = "recognition: %r" % (words,)
            def on_failure(self):
                self._log.debug("SAPI5 RecObs on_failure()")
                self.status = "failure"
        observer = Obs()
        observer.register()

        self._log.debug("SAPI5 mimic: %r" % (words,))
        self.engine.mimic(words)

        timeout = 10
        NULL = c_int(win32con.NULL)
        if timeout != None:
            begin_time = time.time()
            timed_out = False
            windll.user32.SetTimer(NULL, NULL, int(timeout * 1000), NULL)
    
        message = MSG()
        message_pointer = pointer(message)

        while (not timeout) or (time.time() - begin_time < timeout):
            if timeout:
                self._log.debug("SAPI5 message loop: %s sec left" % (timeout + begin_time - time.time()))
            else:
                self._log.debug("SAPI5 message loop: no timeout")

            if windll.user32.GetMessageW(message_pointer, NULL, 0, 0) == 0:
                msg = str(WinError())
                self._log.error("GetMessageW() failed: %s" % msg)
                raise EngineError("GetMessageW() failed: %s" % msg)

            self._log.debug("SAPI5 message: %r" % (message.message,))
            if message.message == win32con.WM_TIMER:
                # A timer message means this loop has timed out.
                self._log.debug("SAPI5 message loop timed out: %s sec left" % (timeout + begin_time - time.time()))
                timed_out = True
                break
            else:
                # Process other messages as normal.
                self._log.debug("SAPI5 message translating and dispatching.")
                windll.user32.TranslateMessage(message_pointer)
                windll.user32.DispatchMessageW(message_pointer)

            if self._recognized_value != self._NotSet:
                # The previous message was a recognition which matched.
                self._log.debug("SAPI5 message caused recognition.")
                break

        observer.unregister()

        if self._recognized_value == self._NotSet:
            if observer.status == "failure":
                raise MimicFailure("Mimic failed.")
            elif observer.status == "none":
                raise MimicFailure("Mimic failed, nothing happened.")

    _mimic_methods["sapi5"] = _mimic_sapi5
    _mimic_methods["sapi5shared"] = _mimic_sapi5




import time
from collections          import defaultdict
from ...grammar.elements  import Compound
import logging


#===========================================================================

class CommandSet(object):

    def __init__(self, commands=None):
        if commands == None:  self._commands = []
        else:                 self._commands = commands

    @property
    def commands(self):
        if self._commands == None:
            return ()
        else:
            return tuple(self._commands)

    def get_poststate_commands_pairs(self):
        mapping = defaultdict(list)
        ids = {}
        for command in self._commands:
            key = id(command.poststate)
            mapping[key].append(command)
            ids[key] = command.poststate
        return [(ids[k], cmds) for (k, cmds) in mapping.items()]

    def add_command(self, command):
        self._commands.append(command)

    def remove_command(self, command):
        if not self._commands:
            return
        try:
            self._commands.remove(command)
        except Exception:
            pass

    def add_commands(self, *commands):
        self._commands.extend(commands)


#===========================================================================

class CommandBase(object):

    _log = logging.getLogger("Command")

    def __init__(self):
        pass

    def __str__(self):
        return "%s()" % (self.__class__.__name__)

    @property
    def element(self):
        raise NotImplementedError("Virtual property %r of class %s."
                                  % ("element", self.__class__.__name__))

    def execute(self, node):
        raise NotImplementedError("Virtual property %r of class %s."
                                  % ("execute", self.__class__.__name__))


#---------------------------------------------------------------------------

class CompoundCommand(CommandBase):

    def __init__(self, spec, action, extras=None):
        CommandBase.__init__(self)
        if extras:  extras = list(extras)
        else:       extras = []
        if isinstance(action, ExtendedAction):
            self._action = action.action
            extras.extend(action.extras)
        else:
            self._action = action
        self._spec = spec
        self._extras = extras
        self._element = Compound(spec, extras=extras)
        CommandBase.__init__(self)

    def __str__(self):
        args = self._spec
        return "%s(%s)" % (self.__class__.__name__, args)

    @property
    def element(self):
        return self._element

    def execute(self, node):
#        self._log.warning("executing CompoundCommand: %s" % node.words())

        extras = {"_node": node}
        for element in self._extras:
            name = element.name
            extra_node = node.get_child_by_name(name, shallow=True)
            if extra_node:
                extras[name] = extra_node.value()
            elif element.has_default():
                extras[name] = element.default

        bound_action = self._action.copy_bind(extras)
        bound_action.execute()


#---------------------------------------------------------------------------

class ExtendedAction(object):
    def __init__(self, action, extras):
        self.action = action
        self.extras = extras



#
# This file is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it 
#   under the terms of the GNU Lesser General Public License as published 
#   by the Free Software Foundation, either version 3 of the License, or 
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but 
#   WITHOUT ANY WARRANTY; without even the implied warranty of 
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU 
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public 
#   License along with Dragonfly.  If not, see 
#   <http://www.gnu.org/licenses/>.
#

"""
SR back-end package for DNS and Natlink
============================================================================

"""

import logging
_log = logging.getLogger("engine.natlink")


#---------------------------------------------------------------------------

# Module level singleton instance of this engine implementation.
_engine = None


def is_engine_available():
    """ Check whether Natlink is available. """
    global _engine
    if _engine:
        return True

    # Attempt to import natlink.
    try:
        import natlink
    except ImportError as e:
        _log.info("Failed to import natlink package: %s" % (e,))
        return False
    except Exception as e:
        _log.exception("Exception during import of natlink package: %s" % (e,))
        return False

    try:
        if natlink.isNatSpeakRunning():
            return True
        else:
            _log.info("Natlink is available but NaturallySpeaking is not"
                      " running.")
            return False
    except Exception as e:
        _log.exception("Exception during natlink.isNatSpeakRunning(): %s" % (e,))
        return False


def get_engine():
    """ Retrieve the Natlink back-end engine object. """
    global _engine
    if not _engine:
        from .engine import NatlinkEngine
        _engine = NatlinkEngine()
    return _engine



#
# This file is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it 
#   under the terms of the GNU Lesser General Public License as published 
#   by the Free Software Foundation, either version 3 of the License, or 
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but 
#   WITHOUT ANY WARRANTY; without even the implied warranty of 
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU 
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public 
#   License along with Dragonfly.  If not, see 
#   <http://www.gnu.org/licenses/>.
#

"""
    This file offers an interface to the Win32 information about
    available monitors (a.k.a. screens, displays).
"""
from six import integer_types

import win32api
import logging

from .rectangle import Rectangle


#---------------------------------------------------------------------------
# List of monitors that will be filled when this module is loaded and
# updated on calls to `Monitor.update_monitors_list`.

monitors = []


#===========================================================================
# Monitor class for storing info about single display monitor.

class Monitor(object):

    _log = logging.getLogger("monitor.init")

    #-----------------------------------------------------------------------
    # Methods for initialization and introspection.

    def __init__(self, handle, rectangle):
        assert isinstance(handle, integer_types)
        self._handle = handle
        assert isinstance(rectangle, Rectangle)
        self._rectangle = rectangle

    def __str__(self):
        return "%s(%d)" % (self.__class__.__name__, self._handle)

    @staticmethod
    def update_monitors_list():
        """
        Update dragonfly's monitors list with new monitors, delete monitors
        that aren't found and replace rectangles of monitors that have
        changed.
        """
        # Get an updated list of monitor information and the handles of
        # monitors in the list at the moment.
        updated_monitors = win32api.EnumDisplayMonitors()
        monitor_handles = [m.handle for m in monitors]

        # Update the list with any new or changed monitors.
        for h1, _, rectangle in updated_monitors:
            handle = h1.handle
            x1, x2, dx, dy = rectangle

            # Check if the monitors list already contains this monitor.
            if handle in monitor_handles:
                # Replace the rectangle for the monitor if it has changed.
                i = monitor_handles.index(handle)
                m = monitors[i]
                r = Rectangle(x1, x2, dx, dy)
                if r != m._rectangle:
                    Monitor._log.debug("Setting %s.rectangle to %s"
                                       % (m, r))
                    m.rectangle = r
            else:
                # Add the new monitor to the list.
                m = Monitor(handle, Rectangle(x1, x2, dx, dy))
                Monitor._log.debug("Adding %s to monitors list" % m)
                monitors.append(m)

        # Remove any monitors that aren't in the updated list.
        monitor_handles = [h1.handle for h1, _, _ in updated_monitors]
        for m in list(monitors):
            if m.handle not in monitor_handles:
                Monitor._log.debug("Removing %s from monitors list" % m)
                monitors.remove(m)

    #-----------------------------------------------------------------------
    # Methods that control attribute access.

    def _set_handle(self, handle):
        assert isinstance(handle, integer_types)
        self._handle = handle
    handle = property(fget=lambda self: self._handle,
                      fset=_set_handle,
                      doc="Protected access to handle attribute.")

    def _get_updated_rectangle(self):
        Monitor.update_monitors_list()
        return self._rectangle

    def _set_rectangle(self, rectangle):
        assert isinstance(rectangle, Rectangle)
        self._rectangle = rectangle  # Should make a copy??
    rectangle = property(fget=_get_updated_rectangle,
                         fset=_set_rectangle,
                         doc="Protected access to rectangle attribute.")


# Enumerate monitors and build Dragonfly's monitor list when this
# module is loaded.
Monitor.update_monitors_list()



#
# This file is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it
#   under the terms of the GNU Lesser General Public License as published
#   by the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public
#   License along with Dragonfly.  If not, see
#   <http://www.gnu.org/licenses/>.
#

"""
RunCommand action
============================================================================

The :class:`RunCommand` action takes a command-line program to run including
any required arguments. On execution, the program will be started as a
subprocess.

This action should work correctly on Windows and other platforms.

Example using the optional function parameter::

    from dragonfly import RunCommand

    def func(proc):
        # Read lines from the process.
        for line in iter(proc.stdout.readline, b''):
            print(line)

    RunCommand('ping localhost', func).execute()


Example using a subclass::

    from dragonfly import RunCommand

    class Ping(RunCommand):
        command = "ping localhost"
        def process_command(self, proc):
            # Read lines from the process.
            for line in iter(proc.stdout.readline, b''):
                print(line)

    Ping().execute()


Class reference
----------------------------------------------------------------------------

"""

import os
import shlex
import subprocess

from six import string_types

from .action_base import ActionBase

# --------------------------------------------------------------------------


class RunCommand(ActionBase):
    """
        Start an application from the command-line.

        This class is similar to the :class:`StartApp` class, but is
        designed for running command-line applications and optionally
        processing subprocesses.

    """
    command = None

    def __init__(self, command=None, process_command=None):
        """
            Constructor arguments:
             - *command* (str) -- the command to run when this action is
               executed. The command may include arguments and will be
               parsed by :meth:`shlex.split`.
             - *process_command* (callable) -- optional callable to invoke
               with the :class:`Popen` object after successfully starting
               the subprocess. Using this argument effectively overrides
               the :meth:`process_command` method.

        """
        ActionBase.__init__(self)

        # Complex handling of arguments because of clashing use of the names
        # at the class level: property & class-value.
        if command is not None:
            self.command = command
        if not (self.command and isinstance(self.command, string_types)):
            raise TypeError("command must be a non-empty string, not %s"
                            % self.command)
        if callable(process_command):
            self.process_command = process_command
        if not callable(self.process_command):
            raise TypeError("process_command must be a callable object or None")

    def process_command(self, proc):
        """
        Virtual method to override for custom handling of the command's
        :class:`Popen` object.
        """

    def _execute(self, data=None):
        self._log.info("Executing: %s" % self.command)

        # Suppress showing the new CMD.exe window on Windows.
        startupinfo = None
        if os.name == 'nt':
            startupinfo = subprocess.STARTUPINFO()
            startupinfo.dwFlags |= subprocess.STARTF_USESHOWWINDOW
        try:
            proc = subprocess.Popen(shlex.split(self.command),
                                    stdout=subprocess.PIPE,
                                    stderr=subprocess.STDOUT,
                                    stdin=subprocess.PIPE,
                                    startupinfo=startupinfo)
        except Exception as e:
            self._log.exception("Exception from starting subprocess %s: %s"
                                % (self.command, e))
            return False

        self.process_command(proc)



#
# This file is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it 
#   under the terms of the GNU Lesser General Public License as published 
#   by the Free Software Foundation, either version 3 of the License, or 
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but 
#   WITHOUT ANY WARRANTY; without even the implied warranty of 
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU 
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public 
#   License along with Dragonfly.  If not, see 
#   <http://www.gnu.org/licenses/>.
#

"""
This file implements a standard Win32 button control.

"""


#---------------------------------------------------------------------------

import win32con

from dragonfly.windows.control_base import ControlBase


#---------------------------------------------------------------------------

class Button(ControlBase):

    def __init__(self, parent, text, size, default=False, **kwargs):
        flavor = 128
        style =  ( win32con.BS_PUSHBUTTON
                 | win32con.BS_TEXT
                 | win32con.WS_CHILD
                 | win32con.WS_TABSTOP
                 | win32con.WS_OVERLAPPED
                 | win32con.WS_VISIBLE)
        if default:  style |= win32con.BS_DEFPUSHBUTTON
        else:        style |= win32con.BS_PUSHBUTTON
        ControlBase.__init__(self, parent, flavor, text, size, style,
                             **kwargs)



import logging

import jsgf
from jsgf import map_expansion, VoidRef
from jsgf.ext import SequenceRule, dictation_in_expansion

from dragonfly import Grammar, List as DragonflyList, DictList as DragonflyDictList
from dragonfly.grammar.elements import *
from dragonfly.grammar.elements_compound import stuff
from dragonfly.grammar.rule_base import Rule
from dragonfly.parser import Parser, ParserError


class TranslationError(Exception):
    pass


class LinkedGrammar(jsgf.Grammar):
    def __init__(self, name, df_grammar):
        self._df_grammar = df_grammar
        super(LinkedGrammar, self).__init__(name=name)

    @property
    def df_grammar(self):
        """
        Dragonfly grammar that this one was built from.
        :return: Grammar
        """
        return self._df_grammar


class LinkedRule(jsgf.Rule):
    def __init__(self, name, visible, expansion, df_rule):
        self._df_rule = df_rule
        super(LinkedRule, self).__init__(name, visible, expansion)

    @property
    def df_rule(self):
        """
        Dragonfly rule that this one was built from.
        :return: Rule
        """
        return self._df_rule


class PatchedRepeat(jsgf.Repeat):
    """
    Repeat class patched to compile JSGF repeats as "expansion [expansion]*" to
    avoid a bug in Pocket Sphinx with the repeat operator.
    """
    def compile(self, ignore_tags=False):
        super(PatchedRepeat, self).compile()
        compiled = self.child.compile(ignore_tags)
        if self.tag and not ignore_tags:
            return "(%s)[%s]*%s" % (compiled, compiled, self.tag)
        else:
            return "(%s)[%s]*" % (compiled, compiled)


class TranslationState(object):
    """
    Translation state used in building JSGF expansions, rules and grammars from
    dragonfly elements, rules and grammars.
    """
    def __init__(self, element, expansion=None, dependencies=None):
        """
        :type element: ElementBase
        """
        self.element = element
        self._expansion = expansion
        if isinstance(dependencies, list):
            self.dependencies = dependencies
        else:
            self.dependencies = []

        # Used if this object is being used to translate a dragonfly rule.
        self.jsgf_rule = None

    rule_names = property(lambda self: [rule.name for rule in self.dependencies])

    @property
    def expansion(self):
        return self._expansion

    @expansion.setter
    def expansion(self, value):
        # Valid values are an Expansion object or None
        if value is not None and not isinstance(value, jsgf.Expansion):
            raise AttributeError("value must be an Expansion object or None.")

        self._expansion = value


class Translator(object):
    _log = logging.getLogger("dragonfly2jsgf")
    _parser = Parser(stuff, _log)

    def __init__(self):
        # Used to not add duplicate rules in translate_grammar that were already
        # translated in translate_repetition_with_dictation.
        self.expansion_black_list = []

    def translate(self, df_spec):
        element = self._parser.parse(df_spec)
        if not element:
            raise ParserError("Invalid dragonfly spec: %r" % df_spec)

        assert isinstance(element, ElementBase)
        state = TranslationState(element)
        return self.get_jsgf_equiv(state)

    def translate_grammar(self, grammar):
        """
        Translate a dragonfly grammar into a JSGF grammar.
        :type grammar: Grammar
        :return: LinkedGrammar
        """
        self.expansion_black_list = []
        result = LinkedGrammar(grammar.name, grammar)
        for rule in grammar.rules:
            state = self.translate_rule(rule)

            # Add the translated rule and any dependencies found during translation
            result.add_rule(state.jsgf_rule)
            for d in state.dependencies:
                if d not in result.rules:  # don't add the same rule twice
                    result.add_rule(d)

        # Remove any rules with black listed expansions added during translation of
        # repeating dictation rules.
        for rule in result.rules:
            if rule.expansion in self.expansion_black_list:
                result.remove_rule(rule)
        return result

    def translate_rule(self, rule):
        """
        Translate a dragonfly rule into a JSGF rule.
        :type rule: Rule
        :return: TranslationState
        """
        element = rule.element
        state = self.get_jsgf_equiv(TranslationState(element))
        state.jsgf_rule = LinkedRule(rule.name, rule.exported, state.expansion,
                                     rule)
        return state

    def translate_rule_ref(self, state):
        """
        Translate the dragonfly RuleRef object inside the TranslationState
        object by collecting all of the rules as necessary using the
        'dependencies' list.
        :type state: TranslationState
        :return TranslationState
        """
        element = state.element
        if not isinstance(element, RuleRef):
            raise TranslationError("Cannot translate element '%s' as a "
                                   "RuleRef." % state.element)
        name = element.name
        if name is None: name = element.element_tree_string()  # f"RuleRef({referenced_rule_name})"
        if name not in state.rule_names:
            # Make a new equivalent JSGF rule
            state.element = element.rule.element
            self.get_jsgf_equiv(state)
            rule_expansion = state.expansion
            new_rule = jsgf.Rule(name, element.rule.exported, rule_expansion)
            state.element = element
            state.dependencies.append(new_rule)
            state.expansion = jsgf.RuleRef(new_rule)
        else:
            # Use the existing rule
            index = state.rule_names.index(name)
            state.expansion = jsgf.RuleRef(state.dependencies[index])

        return state

    def translate_to_rule(self, rule_name, df_spec, visible=True):
        expansion = self.translate(df_spec).expansion
        return jsgf.Rule(rule_name, visible, expansion)

    @staticmethod
    def translate_list(lst):
        """
        Translate a dragonfly List object into a JSGF rule representing it.
        The returned rule is a HiddenRule because lists can only be referenced by
        rules using ListRef, not spoken directly.
        :type lst: DragonflyList
        :return: HiddenRule
        """
        return jsgf.HiddenRule(lst.name, jsgf.AlternativeSet(*lst))

    @staticmethod
    def translate_dict_list(dict_list):
        """
        Translate a dragonfly DictList object into a JSGF rule representing it.
        The returned rule is a HiddenRule because dict lists can only be referenced
        by rules using DictListRef, not spoken directly.
        :type dict_list: DragonflyDictList
        :return: HiddenRule
        """
        keys = list(dict_list.keys())
        keys.sort()
        return jsgf.HiddenRule(dict_list.name, jsgf.AlternativeSet(*keys))

    @staticmethod
    def translate_list_ref(state):
        """
        Translate the dragonfly ListRef element inside the state object by
        creating a JSGF rule representing the List.

        JSGF has no equivalent for Lists or DictLists, however in the spirit
        of making this engine as full-featured as possible, they can be
        implemented as JSGF rules.

        state.expansion will be set to a JSGF RuleRef for the new rule and the new
        rule itself will be added to the state.dependencies list.

        :type state: TranslationState
        :return: TranslationState
        """
        element = state.element
        if not isinstance(element, (List, ListRef)):
            raise TranslationError("Cannot translate element '%s' as a "
                                   "ListRef." % element)

        # Make a new JSGF rule representing the list and set the appropriate values
        # in the state object
        list_rule = Translator.translate_list(element.list)
        state.dependencies.append(list_rule)
        state.expansion = jsgf.RuleRef(list_rule)
        return state

    @staticmethod
    def translate_dict_list_ref(state):
        """
        Translate the dragonfly DictListRef element inside the state object by
        creating a JSGF rule representing the DictList.

        state.expansion will be set to a JSGF RuleRef for the new rule and the new
        rule itself will be added to the state.dependencies list.

        :type state: TranslationState
        :return: TranslationState
        """
        element = state.element
        if not isinstance(element, (DictList, DictListRef)):
            raise TranslationError("Cannot translate element '%s' as a "
                                   "DictListRef." % element)

        dict_list_rule = Translator.translate_dict_list(element.list)
        state.dependencies.append(dict_list_rule)
        state.expansion = jsgf.RuleRef(dict_list_rule)
        return state

    def _has_dictation_descendant(self, element):
        """
        Find if there is a Dictation element in an element tree. This will also
        find Dictation elements in any referenced rule's tree.
        :return: list
        """
        result = False
        if isinstance(element, Dictation):
            result = True
        elif isinstance(element, RuleRef):
            result = result or self._has_dictation_descendant(element.rule.element)
        else:
            for child in element.children:
                result = result or self._has_dictation_descendant(child)
        return result

    def translate_repetition_with_dictation(self, state):
        """
        Translate a dragonfly Repetition element that has at least one Dictation
        descendant.
        :type state: TranslationState
        :return: TranslationState
        """
        element = state.element

        # Calculate the equivalent JSGF expansion for the first child and ignore the
        # rest (if any); such children are used for min/max limits on the number of
        # repetitions. These limits doesn't make much sense for Dictation
        # functionality that has required utterance pauses. Any min/max limit is
        # handled by the engine on a JSGF rule level instead of here.
        state.element = state.element.children[0]
        self.get_jsgf_equiv(state)
        child = state.expansion

        # Switch element back to the Repetition
        state.element = element

        # Process rule references in the expansion tree with dictation
        def process(x):
            if isinstance(x, jsgf.RuleRef) and dictation_in_expansion(x):
                # Skip the RuleRef and use the rule's expansion tree instead
                if x.parent:
                    # If there is a parent, replace x with the referenced rule's
                    # expansion in the parent's children list.
                    i = x.parent.children.index(x)
                    x.parent.children.remove(x)
                    x.parent.children.insert(i, x.rule.expansion)
                x.referenced_rule.expansion.parent = x.parent

                # The rule is no longer a dependency and should not be included in
                # the final grammar
                state.dependencies.remove(x.referenced_rule)

                # In case it was translated before this method, blacklist it!
                self.expansion_black_list.append(x.referenced_rule.expansion)

                # Replace the original child if necessary
                if x is child:
                    state.expansion = x.referenced_rule.expansion

        map_expansion(child, process)

        # Wrap the new expansion in a Repeat
        state.expansion = PatchedRepeat(state.expansion)
        return state

    def get_jsgf_equiv(self, state):
        """
        Take a TranslationState object containing a dragonfly element
        return the same object with the 'expansion' member set to the
        equivalent JSGF expansion object.
        :type state: TranslationState
        :return TranslationState
        """
        element = state.element

        def get_equiv_children():
            children = []
            for child in state.element.children:
                state.element = child
                self.get_jsgf_equiv(state)
                children.append(state.expansion)
            return children

        if isinstance(element, Literal):  # Literal == word list
            text = " ".join(element.words)
            state.expansion = jsgf.Literal(text)

        elif isinstance(element, (RuleRef, Rule)):
            # Translate the dragonfly rule reference
            self.translate_rule_ref(state)

        # DictListRef should be checked before ListRef because it is a subclass
        elif isinstance(element, (DictListRef, DictList)):
            self.translate_dict_list_ref(state)

        elif isinstance(element, (ListRef, List)):
            self.translate_list_ref(state)

        elif isinstance(element, Dictation):
            # A Sphinx decoder for handling dictation will be prepared and used
            # when necessary.
            state.expansion = jsgf.ext.Dictation()

        elif isinstance(element, Impossible):
            # VoidRef is the equivalent of dragonfly's Impossible element.
            # It is also automatically used by pyjsgf if an expansion that should
            # have at least one child (e.g. Sequence/Repeat) doesn't have any at
            # compile time.
            state.expansion = VoidRef()

        elif element.children == ():  # improbable ElementBase case
            state.expansion = jsgf.Expansion([])

        # Repetition should be checked before Sequence because it is a subclass
        elif isinstance(element, Repetition):
            # Repetition should have at least one child
            assert len(state.element.children) > 0

            if self._has_dictation_descendant(element):
                self.translate_repetition_with_dictation(state)
            else:
                # Translate the first child only. The second child is of the form:
                # [first_child [first_child ... first_child ] ... ]
                # This is how matching a limited number of repetitions is done.
                # Limitation (min/max) values for Repetition elements are ignored
                # for the Sphinx engine for the moment.
                equiv_child = get_equiv_children()[0]
                state.expansion = PatchedRepeat(equiv_child)

        elif isinstance(element, Sequence):
            state.expansion = jsgf.Sequence(*get_equiv_children())

        elif isinstance(element, Alternative):
            if len(state.element.children) == 1:
                # Skip redundant (1 child) alternatives
                state.element = state.element.children[0]
                self.get_jsgf_equiv(state)
            else:
                state.expansion = jsgf.AlternativeSet(*get_equiv_children())

        elif isinstance(element, Optional):
            equiv_children = get_equiv_children()
            if len(equiv_children) != 1:
                raise TranslationError("Optional grouping may only have 1 child.")
            state.expansion = jsgf.OptionalGrouping(equiv_children[0])

        return state



# -*- encoding: utf-8 -*-
#
# This file is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it 
#   under the terms of the GNU Lesser General Public License as published 
#   by the Free Software Foundation, either version 3 of the License, or 
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but 
#   WITHOUT ANY WARRANTY; without even the implied warranty of 
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU 
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public 
#   License along with Dragonfly.  If not, see 
#   <http://www.gnu.org/licenses/>.
#

"""
Test cases for the dictation container class for Natlink
============================================================================

"""

import unittest
from six import string_types, text_type

from dragonfly import *
from ..test import TestError, RecognitionFailure, ElementTester
from ..engines.base import DictationContainerBase
from ..engines.backend_natlink.dictation import NatlinkDictationContainer


#===========================================================================

class DictationTestCase(unittest.TestCase):
    """
        Unit testing class for dictation containers.

    """

    input_output = None
    engine = None
    language = None

    #-----------------------------------------------------------------------

    def test_element(self):
        if not self.input_output:
            return
        engine = get_engine()
        if engine.name != self.engine:
            return
        if engine.language != self.language:
            return

        element = Dictation("dictation")
        tester = ElementTester(element)

        failures = []
        try:
            for test_info in self.input_output:
                if len(test_info) == 2:
                    words, expected_value = test_info
                elif len(test_info) == 3:
                    words, expected_value, flags = test_info
                else:
                    raise TestError("Invalid test info: %s" % (test_info,))

                if isinstance(words, string_types):
                    words = words.split()

                recognized_value = tester.recognize(words)
                if isinstance(recognized_value, string_types):
                    recognized_value = text_type(recognized_value)
                elif isinstance(recognized_value, DictationContainerBase):
                    recognized_value = text_type(recognized_value)
                print("result:", recognized_value)
                if recognized_value != expected_value:
                    failures.append((words, expected_value,
                                     recognized_value))
        except TestError as e:
            self.fail(str(e))

        if failures:
            message = ["Recognition mismatches:"]
            for (words, expected_value, recognized_value) in failures:
                message.append("  expected %r, recognized %r (words: %r)"
                               % (expected_value, recognized_value, words))
            self.fail("\n".join(message))


#===========================================================================

class EnglishNatlinkDictationTestCase(DictationTestCase):
    engine       = "natlink"
    language     = "en"
    input_output = [
                    # Trivial words.
                    (u"non-existent-word",           RecognitionFailure),
                    (u"hello",                       u"hello"),
                    (u"hello world",                 u"hello world"),

                    # Capitalization.
                    (u"\\Cap hello",                  u"Hello"),
                    (u"hello \\Cap",                  u"hello"),
                    (u"\\Cap hello world",            u"Hello world"),
                    (u"hello \\Cap world",            u"hello World"),
                    (u"\\Caps-On hello world",        u"Hello World"),
                    (u"\\Caps-Off hello world",       u"hello world"),
                    (u"\\Caps-On hello \\Caps-Off world", u"Hello world"),
                    (u"\\All-Caps hello world",       u"HELLO world"),

                    # Spacing.
                    (u"\\No-Space hello",             u"hello"),
                    (u"hello \\No-Space world",       u"helloworld"),
                    (u"hello \\No-Space \\Cap world",  u"helloWorld"),
                    (u"hello \\Cap \\No-Space world",  u"helloWorld"),
                    (u"\\No-Space-On hello world",    u"helloworld"),
                    (u"\\No-Space-On hello \\No-Space-Off world", u"hello world"),

                    # Words with special formatting.
                    (u".\\full-stop hello world",     u".  Hello world"),
                    (u"hello .\\full-stop world",     u"hello.  World"),
                    (u"hello world .\\full-stop",     u"hello world."),
                    (u",\\comma hello world",         u", hello world"),
                    (u"hello ,\\comma world",         u"hello, world"),
                    (u"hello world ,\\comma",         u"hello world,"),
                    (u"(\\left-paren hello world",    u"(hello world"),
                    (u"hello (\\left-paren world",    u"hello (world"),
                    (u"hello world (\\left-paren",    u"hello world ("),
                    (u"-\\hyphen hello world",        u"-hello world"),
                    (u"hello -\\hyphen world",        u"hello-world"),
                    (u"hello world -\\hyphen",        u"hello world-"),
                    (u"four .\\point seven",          u"4.7"),
                    (u"four .\\dot seven",            u"4.7"),
                    (u"four .\\full-stop seven",      u"4.  7"),

                    # Characters with accents.
                    (u"Düsseldorf",                  u"Düsseldorf"),
                   ]


#===========================================================================

if __name__ == "__main__":
    import sys
    import nose
    loader = nose.loader.TestLoader()
    suite = loader.loadTestsFromModule(sys.modules["__main__"])
    nose.core.TestProgram(suite=suite)



#
# This file is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it 
#   under the terms of the GNU Lesser General Public License as published 
#   by the Free Software Foundation, either version 3 of the License, or 
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but 
#   WITHOUT ANY WARRANTY; without even the implied warranty of 
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU 
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public 
#   License along with Dragonfly.  If not, see 
#   <http://www.gnu.org/licenses/>.
#

"""
Extras factory implementations
============================================================================

"""

import dragonfly
from .loader    import ExtrasFactoryBase, ExtrasFactoryError


#===========================================================================
# Extras factory classes.

class DictationExtrasFactory(ExtrasFactoryBase):

    name = "dictation"

    def build(self, call_info):
        name = call_info.name
        if call_info.arguments:
            raise SyntaxError("Invalid arguments for dictation extra: %r"
                              % (call_info.arguments,))
        return dragonfly.Dictation(name)


#---------------------------------------------------------------------------

class IntegerExtrasFactory(ExtrasFactoryBase):

    name = "integer"

    def build(self, call_info):
        name = call_info.name
        arguments = call_info.arguments
        # Check argument input here.
        min = arguments.pop().value
        max = arguments.pop().value
        if arguments:
            raise SyntaxError("Invalid arguments for integer extra: %r"
                              % (arguments,))
        return dragonfly.Integer(name=name, min=min, max=max)



"""
Utilities for training CMU Pocket Sphinx acoustic models.
"""

import os

import wave
from six import string_types


class TrainingDataWriter(object):
    """
    Class to create .fileids, .transcriptions and .wav files in a specified
    directory.

    If the directory doesn't exist, this class will attempt to create it.
    """
    def __init__(self, data_dir, transcript_name, channels, sample_width, rate):
        """
        :param data_dir: directory where training files should be created.
        :param transcript_name: common file name to use for training files.
        :param channels: number of channels to specify in WAVE header.
        :param sample_width: sample width to specify in the WAVE header.
        :param rate: sample rate to specify in the WAVE header.
        :raises: IOError | OSError
        """
        # Check if data_dir exists and is a directory.
        exists = os.path.exists(data_dir)
        if exists and not os.path.isdir(data_dir):
            raise IOError("'%s' is not a directory. Please use a different file "
                          "file path.")

        # Attempt to make the directory if it doesn't exist. Let the caller handle
        # any errors.
        elif not exists:
            os.mkdir(data_dir)

        # Save parameters for later use.
        self._data_dir = data_dir
        self._transcript_name = transcript_name
        self._channels, self._sample_width = channels, sample_width
        self._rate = rate

        # Open the .fileids and .transcription files (they can exist already)
        fileids_path = os.path.join(data_dir, "%s.fileids" % transcript_name)
        transcriptions_path = os.path.join(data_dir, "%s.transcriptions"
                                           % transcript_name)
        self._fileids_file = open(fileids_path, "a")
        self._transcriptions_file = open(transcriptions_path, "a")

        # Store the next transcript id.
        self._transcript_id = self._get_next_transcript_id()

        # Open a new .wav file.
        self._wav_file = self._get_new_wav_file()

    def _get_next_transcript_id(self):
        """
        Generate a new transcript id based on the relevant .wav files in the data
        directory.
        :rtype: int
        """
        # Find all <transcript_name>.wav files and generate a list of IDs.
        transcript_ids = []
        for f in os.listdir(self._data_dir):
            if f.startswith(self._transcript_name) and f.endswith(".wav"):
                # Strip the transcript name from the start and '.wav' from the end
                # and cast the remaining id string into an integer if it is one.
                id_ = f[len(self._transcript_name):][0:-4]
                if id_.isdigit():
                    transcript_ids.append(int(id_))

        if transcript_ids:
            return max(transcript_ids) + 1
        else:
            return 1

    @property
    def current_wav_filename(self):
        """
        The current .wav filename.
        :rtype: str
        """
        return "%s%04d.wav" % (self._transcript_name, self._transcript_id)

    @property
    def current_absolute_wav_file_path(self):
        return os.path.join(self._data_dir, self.current_wav_filename)

    def _get_new_wav_file(self):
        # Open the .wav file using the lowest filename starting at
        # <transcript_name>0001.wav.
        # Get the next guaranteed transcript ID and absolute file path.
        self._transcript_id = self._get_next_transcript_id()
        wav_path = self.current_absolute_wav_file_path

        # Open a new wave file and write the header.
        # Based on this: https://people.csail.mit.edu/hubert/pyaudio/#record-example
        wf = wave.open(wav_path, "wb")
        wf.setnchannels(self._channels)
        wf.setsampwidth(self._sample_width)
        wf.setframerate(self._rate)
        return wf

    def open_next_wav_file(self):
        """
        Open the next .wav file and write the header information.
        """
        self._wav_file = self._get_new_wav_file()

    def write_to_wave_file(self, buf):
        """
        Method to write an audio buffer to the current .wav file.
        :param buf:
        """
        self._wav_file.writeframes(buf)

    def _discard_current_data(self):
        """
        Internal method to discard the current .wav file.
        """
        self._wav_file.close()
        file_path = self.current_absolute_wav_file_path
        if os.path.exists(file_path):
            os.remove(file_path)

    def finalise(self, hypothesis):
        """
        Method to finish writing training data to audio and text files.
        If passed None or an empty string as a hypothesis, the recording
        will be discarded.
        :param hypothesis: str | None
        """
        if hypothesis is not None and not isinstance(hypothesis, string_types):
            raise TypeError("hypothesis must be a string or None.")

        if not hypothesis:
            self._discard_current_data()
        else:
            # Finish writing the .wav file.
            self._wav_file.close()

            # Get the current .wav filename without the '.wav'.
            transcript_identifier = self.current_wav_filename[0:-4]

            # Append some information to the .fileids and .transcriptions files.
            self._transcriptions_file.write(
                "<s> %s </s> (%s)\n" % (hypothesis, transcript_identifier)
            )
            self._fileids_file.write("%s\n" % transcript_identifier)

    def close_files(self):
        """
        Method to close any open files and discard the current .wav file.
        Should be used in engine.disconnect().
        """
        self._transcriptions_file.close()
        self._fileids_file.close()

        # Also close and discard the current .wav file
        self._discard_current_data()

    def __del__(self):
        # Close any open files if this writer goes out of scope or is deleted.
        self.close_files()



#
# This file is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it 
#   under the terms of the GNU Lesser General Public License as published 
#   by the Free Software Foundation, either version 3 of the License, or 
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but 
#   WITHOUT ANY WARRANTY; without even the implied warranty of 
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU 
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public 
#   License along with Dragonfly.  If not, see 
#   <http://www.gnu.org/licenses/>.
#

"""
.. _RefKey:

Key action
============================================================================

This section describes the :class:`Key` action object.  This 
type of action is used for sending keystrokes to the foreground 
application.  Examples of how to use this class are given in
:ref:`RefKeySpecExamples`.


.. _RefKeySpec:

Keystroke specification format
............................................................................

The *spec* argument passed to the :class:`Key` constructor specifies which 
keystroke events will be emulated.  It is a string consisting of one or 
more comma-separated keystroke elements.  Each of these elements has one 
of the following two possible formats:

Normal press-release key action, optionally repeated several times:
   [*modifiers* ``-``] *keyname* [``/`` *innerpause*] [``:`` *repeat*] [``/`` *outerpause*]

Press-and-hold a key, or release a held-down key:
   [*modifiers* ``-``] *keyname* ``:`` *direction* [``/`` *outerpause*]

The different parts of the keystroke specification are as follows.  Note 
that only *keyname* is required; the other fields are optional.

 - *modifiers* --
   Modifiers for this keystroke.  These keys are held down
   while pressing the main keystroke.
   Can be zero or more of the following:

    - ``a`` -- alt key
    - ``c`` -- control key
    - ``s`` -- shift key
    - ``w`` -- Windows key

 - *keyname* --
   Name of the keystroke.  Valid names are listed in
   :ref:`RefKeySpecNames`.
 - *innerpause* --
   The time to pause between repetitions of this keystroke.
 - *repeat* --
   The number of times this keystroke should be repeated.
   If not specified, the key will be pressed and released once.
 - *outerpause* --
   The time to pause after this keystroke.
 - *direction* --
   Whether to press-and-hold or release the key.  Must be
   one of the following:

    - ``down`` -- press and hold the key
    - ``up`` -- release the key

   Note that releasing a key which is *not* being held down does *not* 
   cause an error.  It harmlessly does nothing.


.. _RefKeySpecNames:

Key names
............................................................................

 - Lowercase letter keys: ``a`` or ``alpha``, ``b`` or ``bravo``,
   ``c`` or ``charlie``, ``d`` or ``delta``, ``e`` or ``echo``,
   ``f`` or ``foxtrot``, ``g`` or ``golf``, ``h`` or ``hotel``,
   ``i`` or ``india``, ``j`` or ``juliet``, ``k`` or ``kilo``,
   ``l`` or ``lima``, ``m`` or ``mike``, ``n`` or ``november``,
   ``o`` or ``oscar``, ``p`` or ``papa``, ``q`` or ``quebec``,
   ``r`` or ``romeo``, ``s`` or ``sierra``, ``t`` or ``tango``,
   ``u`` or ``uniform``, ``v`` or ``victor``, ``w`` or ``whisky``,
   ``x`` or ``xray``, ``y`` or ``yankee``, ``z`` or ``zulu``
 - Uppercase letter keys: ``A`` or ``Alpha``, ``B`` or ``Bravo``,
   ``C`` or ``Charlie``, ``D`` or ``Delta``, ``E`` or ``Echo``,
   ``F`` or ``Foxtrot``, ``G`` or ``Golf``, ``H`` or ``Hotel``,
   ``I`` or ``India``, ``J`` or ``Juliet``, ``K`` or ``Kilo``,
   ``L`` or ``Lima``, ``M`` or ``Mike``, ``N`` or ``November``,
   ``O`` or ``Oscar``, ``P`` or ``Papa``, ``Q`` or ``Quebec``,
   ``R`` or ``Romeo``, ``S`` or ``Sierra``, ``T`` or ``Tango``,
   ``U`` or ``Uniform``, ``V`` or ``Victor``, ``W`` or ``Whisky``,
   ``X`` or ``Xray``, ``Y`` or ``Yankee``, ``Z`` or ``Zulu``
 - Number keys: ``0`` or ``zero``, ``1`` or ``one``, ``2`` or ``two``,
   ``3`` or ``three``, ``4`` or ``four``, ``5`` or ``five``,
   ``6`` or ``six``, ``7`` or ``seven``, ``8`` or ``eight``,
   ``9`` or ``nine``
 - Symbol keys: ``bang`` or ``exclamation``, ``at``, ``hash``,
   ``dollar``, ``percent``, ``caret``, ``and`` or ``ampersand``,
   ``star`` or ``asterisk``, ``leftparen`` or ``lparen``,
   ``rightparen`` or ``rparen``, ``minus`` or ``hyphen``,
   ``underscore``, ``plus``, ``backtick``, ``tilde``,
   ``leftbracket`` or ``lbracket``, ``rightbracket`` or ``rbracket``,
   ``leftbrace`` or ``lbrace``, ``rightbrace`` or ``rbrace``,
   ``backslash``, ``bar``, ``colon``, ``semicolon``,
   ``apostrophe`` or ``singlequote`` or ``squote``,
   ``quote`` or ``doublequote`` or ``dquote``, ``comma``, ``dot``,
   ``slash``, ``lessthan`` or ``leftangle`` or ``langle``,
   ``greaterthan`` or ``rightangle`` or ``rangle``, ``question``,
   ``equal`` or ``equals``
 - Whitespace and editing keys: ``enter``, ``tab``, ``space``,
   ``backspace``, ``delete`` or ``del``
 - Modifier keys: ``shift``, ``control`` or ``ctrl``, ``alt``
 - Special keys: ``escape``, ``insert``, ``pause``, ``win``,
   ``apps`` or ``popup``
 - Navigation keys: ``up``, ``down``, ``left``, ``right``,
   ``pageup`` or ``pgup``, ``pagedown`` or ``pgdown``, ``home``, ``end``
 - Number pad keys: ``npmul``, ``npadd``, ``npsep``, ``npsub``,
   ``npdec``, ``npdiv``, ``numpad0`` or ``np0``, ``numpad1`` or ``np1``,
   ``numpad2`` or ``np2``, ``numpad3`` or ``np3``,
   ``numpad4`` or ``np4``, ``numpad5`` or ``np5``,
   ``numpad6`` or ``np6``, ``numpad7`` or ``np7``,
   ``numpad8`` or ``np8``, ``numpad9`` or ``np9``
 - Function keys: ``f1``, ``f2``, ``f3``, ``f4``, ``f5``, ``f6``,
   ``f7``, ``f8``, ``f9``, ``f10``, ``f11``, ``f12``, ``f13``, ``f14``,
   ``f15``, ``f16``, ``f17``, ``f18``, ``f19``, ``f20``, ``f21``,
   ``f22``, ``f23``, ``f24``
 - Multimedia keys: ``volumeup`` or ``volup``,
   ``volumedown`` or ``voldown``, ``volumemute`` or ``volmute``,
   ``tracknext``, ``trackprev``, ``playpause``, ``browserback``,
   ``browserforward``


.. _RefKeySpecExamples:

Example key actions
............................................................................

The following code types the text "Hello world!" into the foreground 
application: ::

    Key("H, e, l, l, o, space, w, o, r, l, d, exclamation").execute()

The following code is a bit more useful, as it saves the current file with 
the name "dragonfly.txt" (this works for many English-language 
applications): ::

    action = Key("a-f, a/50") + Text("dragonfly.txt") + Key("enter")
    action.execute()

The following code selects the next four lines by holding down the *shift* 
key, slowly moving down 4 lines, and then releasing the *shift* key: ::

    Key("shift:down, down/25:4, shift:up").execute()

The following code locks the screen by pressing the *Windows* key together 
with the *l*: ::

    Key("w-l").execute()


Key class reference
............................................................................

"""


from .action_base  import DynStrActionBase, ActionError
from .typeables    import typeables
from .keyboard     import Keyboard


#---------------------------------------------------------------------------

class Key(DynStrActionBase):

    """
        Keystroke emulation action.

        Constructor arguments:
         - *spec* (*str*) -- keystroke specification
         - *static* (boolean) -- flag indicating whether the
           specification contains dynamic elements

        The format of the keystroke specification *spec* is described in 
        :ref:`RefKeySpec`.

        This class emulates keyboard activity by sending keystrokes to the 
        foreground application.  It does this using Dragonfly's keyboard 
        interface implemented in the :mod:`keyboard` and :mod:`sendinput` 
        modules.  These use the ``sendinput()`` function of the Win32 API.

    """

    # Various keystroke specification format parameters.
    _key_separator = ","
    _delimiter_characters = ":/"
    _modifier_prefix_delimiter = "-"
    _modifier_prefix_characters = {
        'a': typeables["alt"],
        'c': typeables["control"],
        's': typeables["shift"],
        'w': typeables["win"],
        }
    interval_factor = 0.01
    interval_default = 0.0
    _keyboard = Keyboard()

    def _parse_spec(self, spec):
        # Iterate through the keystrokes specified in spec, parsing
        #  each individually.
        events = []
        for single in spec.split(self._key_separator):
            events.extend(self._parse_single(single))
        return events

    def _parse_single(self, spec):

        # Remove leading and trailing whitespace.
        spec = spec.strip()
        if not spec:
            return []

        # Parse modifier prefix.
        index = spec.find(self._modifier_prefix_delimiter)
        if index != -1:
            s = spec[:index]
            index += 1
            modifiers = []
            for c in s:
                if c not in self._modifier_prefix_characters:
                    raise ActionError("Invalid modifier"
                                      " prefix character: %r" % c)
                m = self._modifier_prefix_characters[c]
                if m in modifiers:
                    raise ActionError("Double modifier"
                                      " prefix character: %r" % c)
                modifiers.append(m)
        else:
            index = 0
            modifiers = ()

        inner_pause = None
        special = None
        outer_pause = None

        delimiters = [(c, i + index) for i, c in enumerate(spec[index:])
                                     if c in self._delimiter_characters]
        delimiter_sequence = "".join([d[0] for d in delimiters])
        delimiter_index = [d[1] for d in delimiters]

        if delimiter_sequence == "":            
            keyname = spec[index:]
        elif delimiter_sequence == "/":
            keyname = spec[index:delimiter_index[0]]
            outer_pause = spec[delimiter_index[0]+1:]
        elif delimiter_sequence == "/:":
            keyname = spec[index:delimiter_index[0]]
            inner_pause = spec[delimiter_index[0]+1:delimiter_index[1]]
            special = spec[delimiter_index[1]+1:]
        elif delimiter_sequence == "/:/":
            keyname = spec[index:delimiter_index[0]]
            inner_pause = spec[delimiter_index[0]+1:delimiter_index[1]]
            special = spec[delimiter_index[1]+1:delimiter_index[2]]
            outer_pause = spec[delimiter_index[2]+1:]
        elif delimiter_sequence == ":":
            keyname = spec[index:delimiter_index[0]]
            special = spec[delimiter_index[0]+1:]
        elif delimiter_sequence == ":/":
            keyname = spec[index:delimiter_index[0]]
            special = spec[delimiter_index[0]+1:delimiter_index[1]]
            outer_pause = spec[delimiter_index[1]+1:]
        else:
            raise ActionError("Invalid key spec: %s" % spec)

        try:
            code = typeables[keyname]
        except KeyError:
            raise ActionError("Invalid key name: %r" % keyname)


        if inner_pause is not None:
            s = inner_pause
            try:
                inner_pause = float(s) * self.interval_factor
                if inner_pause < 0: raise ValueError
            except ValueError:
                raise ActionError("Invalid inner pause value: %r,"
                                  " should be a positive number." % s)
        if outer_pause is not None:
            s = outer_pause
            try:
                outer_pause = float(s) * self.interval_factor
                if outer_pause < 0: raise ValueError
            except ValueError:
                raise ActionError("Invalid outer pause value: %r,"
                                  " should be a positive number." % s)
        else:
            outer_pause = self.interval_default * self.interval_factor

        direction = None; repeat = 1
        if special is not None:
            if special == "down":   direction = True
            elif special == "up":   direction = False
            else:
                s = special
                try:
                    repeat = int(s)
                    if repeat < 0: raise ValueError
                except ValueError:
                    raise ActionError("Invalid repeat value: %r,"
                                      " should be a positive integer." % s)

        if direction is None:
            if inner_pause is None:
                inner_pause = self.interval_default * self.interval_factor
            if repeat == 0:
                events = []
            else:
                events = []
                for m in modifiers:
                    events.extend(m.on_events())
                for i in range(repeat - 1):
                    events.extend(code.events(inner_pause))
                events.extend(code.events(outer_pause))
                for m in modifiers[-1::-1]:
                    events.extend(m.off_events())
        else:
            if modifiers:
                raise ActionError("Cannot use direction with modifiers.")
            if inner_pause is not None:
                raise ActionError("Cannot use direction with inner pause.")
            if direction:
                events = code.on_events(outer_pause)
            else:
                events = code.off_events(outer_pause)

        return events

    def _execute_events(self, events):
        self._keyboard.send_keyboard_events(events)
        return True
